from __future__ import annotations

__all__ = [
    "Abs",
    "Acosh",
    "Add",
    "AddScalar",
    "Arange",
    "Asinh",
    "AsinhUniform",
    "Atanh",
    "AutoRegressive",
    "BaseSequenceGenerator",
    "BaseWrapperSequenceGenerator",
    "Cat2",
    "Cauchy",
    "Clamp",
    "Constant",
    "Cosh",
    "Cumsum",
    "Div",
    "Exp",
    "Exponential",
    "Float",
    "Fmod",
    "Full",
    "HalfCauchy",
    "HalfNormal",
    "Linear",
    "Log",
    "LogNormal",
    "LogUniform",
    "Long",
    "Mul",
    "MulScalar",
    "Multinomial",
    "MultinomialChoice",
    "Neg",
    "Normal",
    "Periodic",
    "Poisson",
    "RandAsinhUniform",
    "RandCauchy",
    "RandExponential",
    "RandHalfCauchy",
    "RandHalfNormal",
    "RandInt",
    "RandLogNormal",
    "RandLogUniform",
    "RandNormal",
    "RandPoisson",
    "RandTruncCauchy",
    "RandTruncExponential",
    "RandTruncHalfCauchy",
    "RandTruncHalfNormal",
    "RandTruncLogNormal",
    "RandTruncNormal",
    "RandUniform",
    "RandWienerProcess",
    "SineWave",
    "Sinh",
    "Sort",
    "Sqrt",
    "Sub",
    "Tanh",
    "TensorSequence",
    "Time",
    "TruncCauchy",
    "TruncExponential",
    "TruncHalfCauchy",
    "TruncHalfNormal",
    "TruncLogNormal",
    "TruncNormal",
    "Uniform",
    "UniformCategorical",
    "is_sequence_generator_config",
    "setup_sequence_generator",
]

from startorch.sequence.ar import AutoRegressiveSequenceGenerator as AutoRegressive
from startorch.sequence.base import (
    BaseSequenceGenerator,
    is_sequence_generator_config,
    setup_sequence_generator,
)
from startorch.sequence.categorical import MultinomialSequenceGenerator as Multinomial
from startorch.sequence.categorical import (
    UniformCategoricalSequenceGenerator as UniformCategorical,
)
from startorch.sequence.cauchy import CauchySequenceGenerator as Cauchy
from startorch.sequence.cauchy import RandCauchySequenceGenerator as RandCauchy
from startorch.sequence.cauchy import (
    RandTruncCauchySequenceGenerator as RandTruncCauchy,
)
from startorch.sequence.cauchy import TruncCauchySequenceGenerator as TruncCauchy
from startorch.sequence.choice import (
    MultinomialChoiceSequenceGenerator as MultinomialChoice,
)
from startorch.sequence.constant import ConstantSequenceGenerator as Constant
from startorch.sequence.constant import FullSequenceGenerator as Full
from startorch.sequence.dtype import FloatSequenceGenerator as Float
from startorch.sequence.dtype import LongSequenceGenerator as Long
from startorch.sequence.exponential import ExponentialSequenceGenerator as Exponential
from startorch.sequence.exponential import (
    RandExponentialSequenceGenerator as RandExponential,
)
from startorch.sequence.exponential import (
    RandTruncExponentialSequenceGenerator as RandTruncExponential,
)
from startorch.sequence.exponential import (
    TruncExponentialSequenceGenerator as TruncExponential,
)
from startorch.sequence.halfcauchy import HalfCauchySequenceGenerator as HalfCauchy
from startorch.sequence.halfcauchy import (
    RandHalfCauchySequenceGenerator as RandHalfCauchy,
)
from startorch.sequence.halfcauchy import (
    RandTruncHalfCauchySequenceGenerator as RandTruncHalfCauchy,
)
from startorch.sequence.halfcauchy import (
    TruncHalfCauchySequenceGenerator as TruncHalfCauchy,
)
from startorch.sequence.halfnormal import HalfNormalSequenceGenerator as HalfNormal
from startorch.sequence.halfnormal import (
    RandHalfNormalSequenceGenerator as RandHalfNormal,
)
from startorch.sequence.halfnormal import (
    RandTruncHalfNormalSequenceGenerator as RandTruncHalfNormal,
)
from startorch.sequence.halfnormal import (
    TruncHalfNormalSequenceGenerator as TruncHalfNormal,
)
from startorch.sequence.joining import Cat2SequenceGenerator as Cat2
from startorch.sequence.linear import LinearSequenceGenerator as Linear
from startorch.sequence.lognormal import LogNormalSequenceGenerator as LogNormal
from startorch.sequence.lognormal import RandLogNormalSequenceGenerator as RandLogNormal
from startorch.sequence.lognormal import (
    RandTruncLogNormalSequenceGenerator as RandTruncLogNormal,
)
from startorch.sequence.lognormal import (
    TruncLogNormalSequenceGenerator as TruncLogNormal,
)
from startorch.sequence.math import AbsSequenceGenerator as Abs
from startorch.sequence.math import AddScalarSequenceGenerator as AddScalar
from startorch.sequence.math import AddSequenceGenerator as Add
from startorch.sequence.math import ClampSequenceGenerator as Clamp
from startorch.sequence.math import CumsumSequenceGenerator as Cumsum
from startorch.sequence.math import DivSequenceGenerator as Div
from startorch.sequence.math import ExpSequenceGenerator as Exp
from startorch.sequence.math import FmodSequenceGenerator as Fmod
from startorch.sequence.math import LogSequenceGenerator as Log
from startorch.sequence.math import MulScalarSequenceGenerator as MulScalar
from startorch.sequence.math import MulSequenceGenerator as Mul
from startorch.sequence.math import NegSequenceGenerator as Neg
from startorch.sequence.math import SqrtSequenceGenerator as Sqrt
from startorch.sequence.math import SubSequenceGenerator as Sub
from startorch.sequence.normal import NormalSequenceGenerator as Normal
from startorch.sequence.normal import RandNormalSequenceGenerator as RandNormal
from startorch.sequence.normal import (
    RandTruncNormalSequenceGenerator as RandTruncNormal,
)
from startorch.sequence.normal import TruncNormalSequenceGenerator as TruncNormal
from startorch.sequence.periodic import PeriodicSequenceGenerator as Periodic
from startorch.sequence.poisson import PoissonSequenceGenerator as Poisson
from startorch.sequence.poisson import RandPoissonSequenceGenerator as RandPoisson
from startorch.sequence.range import ArangeSequenceGenerator as Arange
from startorch.sequence.sort import SortSequenceGenerator as Sort
from startorch.sequence.tensor import TensorSequenceGenerator as TensorSequence
from startorch.sequence.time import TimeSequenceGenerator as Time
from startorch.sequence.trigo import AcoshSequenceGenerator as Acosh
from startorch.sequence.trigo import AsinhSequenceGenerator as Asinh
from startorch.sequence.trigo import AtanhSequenceGenerator as Atanh
from startorch.sequence.trigo import CoshSequenceGenerator as Cosh
from startorch.sequence.trigo import SinhSequenceGenerator as Sinh
from startorch.sequence.trigo import TanhSequenceGenerator as Tanh
from startorch.sequence.uniform import AsinhUniformSequenceGenerator as AsinhUniform
from startorch.sequence.uniform import LogUniformSequenceGenerator as LogUniform
from startorch.sequence.uniform import (
    RandAsinhUniformSequenceGenerator as RandAsinhUniform,
)
from startorch.sequence.uniform import RandIntSequenceGenerator as RandInt
from startorch.sequence.uniform import RandLogUniformSequenceGenerator as RandLogUniform
from startorch.sequence.uniform import RandUniformSequenceGenerator as RandUniform
from startorch.sequence.uniform import UniformSequenceGenerator as Uniform
from startorch.sequence.wave import SineWaveSequenceGenerator as SineWave
from startorch.sequence.wiener import (
    RandWienerProcessSequenceGenerator as RandWienerProcess,
)
from startorch.sequence.wrapper import BaseWrapperSequenceGenerator



from __future__ import annotations

__all__ = ["check_matplotlib", "is_matplotlib_available"]

from importlib.util import find_spec


def check_matplotlib() -> None:
    r"""Checks if the matplotlib package is installed.

    Raises:
    ------
        RuntimeError if the matplotlib package is not installed.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.utils.imports import check_matplotlib
        >>> check_matplotlib()
    """
    if not is_matplotlib_available():
        raise RuntimeError(
            "`matplotlib` package is required but not installed. "
            "You can install `matplotlib` package with the command:\n\n"
            "pip install matplotlib\n"
        )


def is_matplotlib_available() -> bool:
    r"""Indicates if the NumPy package is installed or not.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.utils.imports import is_matplotlib_available
        >>> is_matplotlib_available()
    """
    return find_spec("matplotlib") is not None



from __future__ import annotations

__all__ = [
    "AbsTensorGenerator",
    "AddScalarTensorGenerator",
    "AddTensorGenerator",
    "ClampTensorGenerator",
    "DivTensorGenerator",
    "ExpTensorGenerator",
    "FmodTensorGenerator",
    "LogTensorGenerator",
    "MulScalarTensorGenerator",
    "MulTensorGenerator",
    "NegTensorGenerator",
    "SqrtTensorGenerator",
    "SubTensorGenerator",
]

from collections.abc import Sequence

from coola.utils.format import str_indent, str_mapping, str_sequence
from torch import Generator, Tensor

from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator
from startorch.tensor.wrapper import BaseWrapperTensorGenerator


class AbsTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a tensor generator that computes the absolute value of
    a tensor.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Abs, RandNormal
        >>> generator = Abs(RandNormal())
        >>> generator
        AbsTensorGenerator(
          (tensor): RandNormalTensorGenerator(mean=0.0, std=1.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).abs()


class AddScalarTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a tensor generator that adds a scalar value to a
    generated batch of tensors.

    Args:
    ----
        tensor (``BaseTensorGenerator`` or dict):
            Specifies the tensor generator or its configuration.
        value (int or float): Specifies the scalar value to add.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import AddScalar, RandUniform
        >>> generator = AddScalar(RandUniform(), 10.0)
        >>> generator
        AddScalarTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=0.0, high=1.0)
          (value): 10.0
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        generator: BaseTensorGenerator | dict,
        value: int | float,
    ) -> None:
        super().__init__(generator=generator)
        self._value = value

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"tensor": self._generator, "value": self._value}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        tensor = self._generator.generate(size=size, rng=rng)
        tensor.add_(self._value)
        return tensor


class AddTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator that adds several tensor.

    ``tensor = tensor_1 + tensor_2 + ... + tensor_n``

    Args:
    ----
        tensors (``Tensor``): Specifies the tensor generators.

    Raises:
    ------
        ValueError if no sequence generator is provided.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Add, RandNormal, RandUniform
        >>> generator = Add([RandUniform(), RandNormal()])
        >>> generator
        AddTensorGenerator(
          (0): RandUniformTensorGenerator(low=0.0, high=1.0)
          (1): RandNormalTensorGenerator(mean=0.0, std=1.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, tensors: Sequence[BaseTensorGenerator | dict]) -> None:
        super().__init__()
        if not tensors:
            raise ValueError(
                "No tensor generator. You need to specify at least one tensor generator"
            )
        self._tensors = tuple(setup_tensor_generator(tensor) for tensor in tensors)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_sequence(self._tensors))}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        output = self._tensors[0].generate(size=size, rng=rng)
        for generator in self._tensors[1:]:
            output.add_(generator.generate(size=size, rng=rng))
        return output


class ClampTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a tensor generator to generate tensors where the
    values are clamped.

    Note: ``min_value`` and ``max_value`` cannot be both ``None``.

    Args:
    ----
        tensor (``BaseTensorGenerator`` or dict):
            Specifies the tensor generator or its configuration.
        min_value (int, float or ``None``): Specifies the lower bound.
            If ``min_value`` is ``None``, there is no lower bound.
        max_value (int, float or ``None``): Specifies the upper bound.
            If ``max_value`` is ``None``, there is no upper bound.

    Raises:
    ------
        ValueError if both ``min`` and ``max`` are ``None``

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Clamp, RandUniform
        >>> generator = Clamp(RandUniform(low=1.0, high=50.0), min_value=2.0, max_value=10.0)
        >>> generator
        ClampTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=1.0, high=50.0)
          (min_value): 2.0
          (max_value): 10.0
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        generator: BaseTensorGenerator | dict,
        min_value: int | float | None,
        max_value: int | float | None,
    ) -> None:
        super().__init__(generator=generator)
        if min_value is None and max_value is None:
            raise ValueError("`min_value` and `max_value` cannot be both None")
        self._min_value = min_value
        self._max_value = max_value

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "tensor": self._generator,
                    "min_value": self._min_value,
                    "max_value": self._max_value,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).clamp(self._min_value, self._max_value)


class DivTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator that divides one tensor by another
    one.

    ``tensor = dividend / divisor`` (a.k.a. true division)

    Args:
    ----
        dividend (``Tensor``): (``BaseTensorGenerator`` or dict):
            Specifies the dividend tensor generator or its
            configuration.
        divisor (``Tensor``): (``BaseTensorGenerator`` or dict):
            Specifies the divisor tensor generator or its
            configuration.
        rounding_mode (str or ``None``, optional): Specifies the
            type of rounding applied to the result.
            - ``None``: true division.
            - ``"trunc"``: rounds the results of the division
                towards zero.
            - ``"floor"``: floor division.
            Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Div, RandUniform
        >>> generator = Div(RandUniform(), RandUniform(low=1.0, high=10.0))
        >>> generator
        DivTensorGenerator(
          (dividend): RandUniformTensorGenerator(low=0.0, high=1.0)
          (divisor): RandUniformTensorGenerator(low=1.0, high=10.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        dividend: BaseTensorGenerator | dict,
        divisor: BaseTensorGenerator | dict,
        rounding_mode: str | None = None,
    ) -> None:
        super().__init__()
        self._dividend = setup_tensor_generator(dividend)
        self._divisor = setup_tensor_generator(divisor)
        self._rounding_mode = rounding_mode

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"dividend": self._dividend, "divisor": self._divisor}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._dividend.generate(size=size, rng=rng).div(
            self._divisor.generate(size=size, rng=rng),
            rounding_mode=self._rounding_mode,
        )


class ExpTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a tensor generator that computes the exponential of a
    tensor.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Exp, RandUniform
        >>> generator = Exp(RandUniform(low=1.0, high=5.0))
        >>> generator
        ExpTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=1.0, high=5.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).exp()


class FmodTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator that computes the element-wise
    remainder of division.

    Args:
    ----
        dividend (``BaseTensorGenerator`` or dict):
            Specifies the tensor generator (or its configuration) that
            generates the dividend values.
        divisor (int or float): Specifies the divisor.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Fmod, RandUniform
        >>> generator = Fmod(dividend=RandUniform(low=-100, high=100), divisor=10.0)
        >>> generator
        FmodTensorGenerator(
          (dividend): RandUniformTensorGenerator(low=-100.0, high=100.0)
          (divisor): 10.0
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
        >>> generator = Fmod(
        ...     dividend=RandUniform(low=-100, high=100), divisor=RandUniform(low=1, high=10)
        ... )
        >>> generator
        FmodTensorGenerator(
          (dividend): RandUniformTensorGenerator(low=-100.0, high=100.0)
          (divisor): RandUniformTensorGenerator(low=1.0, high=10.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        dividend: BaseTensorGenerator | dict,
        divisor: BaseTensorGenerator | dict | int | float,
    ) -> None:
        super().__init__()
        self._dividend = setup_tensor_generator(dividend)
        self._divisor = setup_tensor_generator(divisor) if isinstance(divisor, dict) else divisor

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"dividend": self._dividend, "divisor": self._divisor}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        tensor = self._dividend.generate(size=size, rng=rng)
        divisor = self._divisor
        if isinstance(divisor, BaseTensorGenerator):
            divisor = divisor.generate(size=size, rng=rng)
        tensor.fmod_(divisor)
        return tensor


class LogTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a tensor generator that computes the logarithm of a
    tensor.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Log, RandUniform
        >>> generator = Log(RandUniform(low=1.0, high=100.0))
        >>> generator
        LogTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=1.0, high=100.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).log()


class MulTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator that multiplies multiple tensors.

    ``tensor = tensor_1 * tensor_2 * ... * tensor_n``

    Args:
    ----
        generators (``Sequence``): Specifies the tensor generators.

    Raises:
    ------
        ValueError if no sequence generator is provided.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.tensor import Mul, RandUniform, RandNormal
        >>> generator = Mul([RandUniform(), RandNormal()])
        >>> generator
        MulTensorGenerator(
          (0): RandUniformTensorGenerator(low=0.0, high=1.0)
          (1): RandNormalTensorGenerator(mean=0.0, std=1.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, generators: Sequence[BaseTensorGenerator | dict]) -> None:
        super().__init__()
        if not generators:
            raise ValueError(
                "No tensor generator. You need to specify at least one tensor generator"
            )
        self._tensors = tuple(setup_tensor_generator(generator) for generator in generators)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_sequence(self._tensors))}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        output = self._tensors[0].generate(size=size, rng=rng)
        for generator in self._tensors[1:]:
            output.mul_(generator.generate(size=size, rng=rng))
        return output


class MulScalarTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a tensor generator that multiplies a scalar value to a
    generated batch of tensors.

    Args:
    ----
        tensor (``BaseTensorGenerator`` or dict):
            Specifies the tensor generator or its configuration.
        value (int or float): Specifies the scalar value to multiply.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.tensor import MulScalar, RandUniform, RandNormal
        >>> generator = MulScalar(RandUniform(), 42)
        >>> generator
        MulScalarTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=0.0, high=1.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        generator: BaseTensorGenerator | dict,
        value: int | float,
    ) -> None:
        super().__init__(generator=generator)
        self._value = value

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"tensor": self._generator}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        tensor = self._generator.generate(size=size, rng=rng)
        tensor.mul_(self._value)
        return tensor


class NegTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a tensor generator that computes the negation of a
    generated tensor.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.tensor import Neg, RandNormal
        >>> generator = Neg(RandNormal())
        >>> generator
        NegTensorGenerator(
          (tensor): RandNormalTensorGenerator(mean=0.0, std=1.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return -self._generator.generate(size=size, rng=rng)


class SqrtTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a tensor generator that computes the squared root of a
    tensor.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform, Sqrt
        >>> generator = Sqrt(RandUniform(low=1.0, high=100.0))
        >>> generator
        SqrtTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=1.0, high=100.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).sqrt()


class SubTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator that subtracts two tensors.

    ``tensor = tensor_1 - tensor_2``

    Args:
    ----
        tensor1 (``BaseTensorGenerator`` or dict):
            Specifies the first tensor generator or its
            configuration.
        tensor2 (``BaseTensorGenerator`` or dict):
            Specifies the second tensor generator or its
            configuration.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandNormal, RandUniform, Sub
        >>> generator = Sub(RandUniform(), RandNormal())
        >>> generator
        SubTensorGenerator(
          (tensor1): RandUniformTensorGenerator(low=0.0, high=1.0)
          (tensor2): RandNormalTensorGenerator(mean=0.0, std=1.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self, tensor1: BaseTensorGenerator | dict, tensor2: BaseTensorGenerator | dict
    ) -> None:
        super().__init__()
        self._tensor1 = setup_tensor_generator(tensor1)
        self._tensor2 = setup_tensor_generator(tensor2)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"tensor1": self._tensor1, "tensor2": self._tensor2}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._tensor1.generate(size=size, rng=rng).sub(
            self._tensor2.generate(size=size, rng=rng)
        )



from __future__ import annotations

__all__ = ["MultinomialChoiceTimeSeriesGenerator"]

from collections.abc import Generator, Sequence

import torch
from coola.utils import str_indent
from redcat import BatchDict

from startorch.timeseries.base import (
    BaseTimeSeriesGenerator,
    setup_timeseries_generator,
)
from startorch.utils.format import str_weighted_modules
from startorch.utils.weight import prepare_weighted_generators


class MultinomialChoiceTimeSeriesGenerator(BaseTimeSeriesGenerator):
    r"""Implements a time series generator that select a time series
    generator at each batch.

    This time series generator is used to generate time series with
    different generation processes. The user can specify a list of
    time series generators with an associated weight. The weight is
    used to sample the time series generator with a multinomial
    distribution. Higher weight means that the time series generator
    has a higher probability to be selected at each batch.
    Each dictionary in the ``generators`` input should have the
    following items:

        - a key ``'generator'`` which indicates the time series
            generator or its configuration.
        - an optional key ``'weight'`` with a float value which
            indicates the weight of the time series generator.
            If this key is absent, the weight is set to ``1.0``.

    Args:
    ----
        generators (tuple or list): Specifies the time series
            generators and their weights. See above to learn
            about the expected format.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.timeseries import MultinomialChoice, TimeSeries
        >>> from startorch.sequence import RandUniform, RandNormal
        >>> generator = MultinomialChoice(
        ...     (
        ...         {
        ...             "weight": 2.0,
        ...             "generator": TimeSeries({"value": RandUniform(), "time": RandUniform()}),
        ...         },
        ...         {
        ...             "weight": 1.0,
        ...             "generator": TimeSeries({"value": RandNormal(), "time": RandNormal()}),
        ...         },
        ...     )
        ... )
        >>> generator
        MultinomialChoiceTimeSeriesGenerator(
          (0) [weight=2.0] TimeSeriesGenerator(
              (value): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
              (time): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
            )
          (1) [weight=1.0] TimeSeriesGenerator(
              (value): RandNormalSequenceGenerator(mean=0.0, std=1.0, feature_size=(1,))
              (time): RandNormalSequenceGenerator(mean=0.0, std=1.0, feature_size=(1,))
            )
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        BatchDict(
          (value): tensor([[...]], batch_dim=0, seq_dim=1)
          (time): tensor([[...]], batch_dim=0, seq_dim=1)
        )
    """

    def __init__(self, generators: Sequence[dict[str, BaseTimeSeriesGenerator | dict]]) -> None:
        super().__init__()
        generators, weights = prepare_weighted_generators(generators)
        self._generators = tuple(setup_timeseries_generator(generator) for generator in generators)
        self._weights = torch.as_tensor(weights, dtype=torch.float)

    def __repr__(self) -> str:
        args = str_indent(str_weighted_modules(modules=self._generators, weights=self._weights))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchDict:
        index = torch.multinomial(self._weights, num_samples=1, generator=rng).item()
        return self._generators[index].generate(seq_len=seq_len, batch_size=batch_size, rng=rng)



from __future__ import annotations

__all__ = [
    "BaseExampleGenerator",
    "Hypercube",
    "TimeSeries",
    "is_example_generator_config",
    "setup_example_generator",
]

from startorch.example.base import (
    BaseExampleGenerator,
    is_example_generator_config,
    setup_example_generator,
)
from startorch.example.hypercube import HypercubeExampleGenerator as Hypercube
from startorch.example.timeseries import TimeSeriesExampleGenerator as TimeSeries



from __future__ import annotations

__all__ = ["SortSequenceGenerator"]


from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator
from startorch.sequence.wrapper import BaseWrapperSequenceGenerator


class SortSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that sorts a generated sequence.

    Args:
    ----
        sequence (``BaseSequenceGenerator`` or dict):
            Specifies the sequence generator or its configuration.
        descending (bool, optional): Controls the sorting order.
            If ``True``, the elements are sorted in
            descending order by value. Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import RandUniform, Sort
        >>> generator = Sort(RandUniform())
        >>> generator
        SortSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        generator: BaseSequenceGenerator | dict,
        descending: bool = False,
    ) -> None:
        super().__init__(generator)
        self._descending = bool(descending)

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(
            seq_len=seq_len, batch_size=batch_size, rng=rng
        ).sort_along_seq(self._descending)[0]



from __future__ import annotations

__all__ = ["FullTensorGenerator"]


import torch
from torch import Generator, Tensor

from startorch.tensor.base import BaseTensorGenerator


class FullTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator that fills the tensor with a given
    value.

    Args:
    ----
        value (bool or int or float): Specifies the fill value.
        dtype (``torch.dtype`` or ``None``): Specifies the target
            dtype. ``None`` means the data type is infered from the
            value type. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Full
        >>> generator = Full(value=42)
        >>> generator
        FullTensorGenerator(value=42, dtype=None)
        >>> generator.generate((2, 6))
        tensor([[42, 42, 42, 42, 42, 42],
                [42, 42, 42, 42, 42, 42]])
    """

    def __init__(self, value: bool | int | float, dtype: torch.dtype | None = None) -> None:
        super().__init__()
        self._value = value
        self._dtype = dtype

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(value={self._value}, dtype={self._dtype})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return torch.full(size=size, fill_value=self._value, dtype=self._dtype)



from __future__ import annotations

__all__ = [
    "BasePeriodicSequenceGenerator",
    "Repeat",
    "SineWave",
    "is_periodic_sequence_generator_config",
    "setup_periodic_sequence_generator",
]

from startorch.periodic.sequence.base import (
    BasePeriodicSequenceGenerator,
    is_periodic_sequence_generator_config,
    setup_periodic_sequence_generator,
)
from startorch.periodic.sequence.repeat import RepeatPeriodicSequenceGenerator as Repeat
from startorch.periodic.sequence.wave import (
    SineWavePeriodicSequenceGenerator as SineWave,
)



from __future__ import annotations

__all__ = ["Cat2SequenceGenerator"]


from coola.utils.format import str_indent, str_mapping
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator
from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator


class Cat2SequenceGenerator(BaseSequenceGenerator):
    r"""Implements a sequence generator that concatenate two sequences.

    ``sequence = [sequence_1, sequence_2]``

    Args:
    ----
        generator1 (``BaseSequenceGenerator`` or dict): Specifies the
            first sequence generator or its configuration.
        generator2 (``BaseSequenceGenerator`` or dict): Specifies the
            second sequence generator or its configuration.
        changepoint (``BaseTensorGenerator`` or dict): Specifies the
            change point generator or its configuration.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Cat2, RandUniform, RandNormal
        >>> from startorch.tensor import RandInt
        >>> generator = Cat2(
        ...     generator1=RandUniform(), generator2=RandNormal(), changepoint=RandInt(0, 12)
        ... )
        >>> generator
        Cat2SequenceGenerator(
          (generator1): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
          (generator2): RandNormalSequenceGenerator(mean=0.0, std=1.0, feature_size=(1,))
          (changepoint): RandIntTensorGenerator(low=0, high=12)
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        generator1: BaseSequenceGenerator | dict,
        generator2: BaseSequenceGenerator | dict,
        changepoint: BaseTensorGenerator | dict,
    ) -> None:
        super().__init__()
        self._generator1 = setup_sequence_generator(generator1)
        self._generator2 = setup_sequence_generator(generator2)
        self._changepoint = setup_tensor_generator(changepoint)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "generator1": self._generator1,
                    "generator2": self._generator2,
                    "changepoint": self._changepoint,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        changepoint = max(min(int(self._changepoint.generate((1,), rng=rng).item()), seq_len), 0)
        return self._generator1.generate(
            seq_len=changepoint, batch_size=batch_size, rng=rng
        ).cat_along_seq(
            self._generator2.generate(seq_len=seq_len - changepoint, batch_size=batch_size, rng=rng)
        )



from __future__ import annotations

__all__ = [
    "ExponentialTensorGenerator",
    "RandExponentialTensorGenerator",
    "RandTruncExponentialTensorGenerator",
    "TruncExponentialTensorGenerator",
]


from coola.utils.format import str_indent, str_mapping
from torch import Generator, Tensor

from startorch.random import (
    exponential,
    rand_exponential,
    rand_trunc_exponential,
    trunc_exponential,
)
from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator


class ExponentialTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensors by sampling values from an
    Exponential distribution.

    The rates of the Exponential distribution are generated by the
    rate generator. The rate generator should return the rate for each
    value in the sequence.

    Args:
    ----
        rate (``BaseTensorGenerator`` or dict):
            Specifies the rate generator or its configuration.
            The rate generator should return valid rate values.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Exponential, RandUniform
        >>> generator = Exponential(rate=RandUniform(low=1.0, high=100.0))
        >>> generator
        ExponentialTensorGenerator(
          (rate): RandUniformTensorGenerator(low=1.0, high=100.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, rate: BaseTensorGenerator | dict) -> None:
        super().__init__()
        self._rate = setup_tensor_generator(rate)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"rate": self._rate}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return exponential(self._rate.generate(size=size, rng=rng), generator=rng)


class RandExponentialTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate sequences by sampling values from
    an Exponential distribution.

    Args:
    ----
        rate (float, optional): Specifies the rate of the Exponential
            distribution. Default: ``1.0``

    Raises:
    ------
        ValueError if ``rate`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandExponential
        >>> generator = RandExponential(rate=1.0)
        >>> generator
        RandExponentialTensorGenerator(rate=1.0)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, rate: float = 1.0) -> None:
        super().__init__()
        if rate <= 0:
            raise ValueError(f"rate has to be greater than 0 (received: {rate})")
        self._rate = float(rate)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(rate={self._rate})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_exponential(size=size, rate=self._rate, generator=rng)


class RandTruncExponentialTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate sequences by sampling values from
    a truncated Exponential distribution.

    Args:
    ----
        rate (float, optional): Specifies the rate of the Exponential
            distribution. Default: ``1.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``5.0``

    Raises:
    ------
        ValueError if ``rate`` is not a positive number.
        ValueError if ``max_value`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandTruncExponential
        >>> generator = RandTruncExponential(rate=1.0, max_value=3.0)
        >>> generator
        RandTruncExponentialTensorGenerator(rate=1.0, max_value=3.0)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, rate: float = 1.0, max_value: float = 5.0) -> None:
        super().__init__()
        if rate <= 0:
            raise ValueError(f"rate has to be greater than 0 (received: {rate})")
        self._rate = float(rate)
        if max_value <= 0:
            raise ValueError(f"max_value has to be greater than 0 (received: {max_value})")
        self._max_value = float(max_value)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(rate={self._rate}, max_value={self._max_value})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_trunc_exponential(
            size=size,
            rate=self._rate,
            max_value=self._max_value,
            generator=rng,
        )


class TruncExponentialTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate sequence by sampling values from
    an Exponential distribution.

    Args:
    ----
        rate (``BaseTensorGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the rate.
        max_value (``BaseTensorGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            maximum value (excluded).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform, TruncExponential
        >>> generator = TruncExponential(
        ...     rate=RandUniform(low=1.0, high=10.0),
        ...     max_value=RandUniform(low=1.0, high=100.0),
        ... )
        >>> generator
        TruncExponentialTensorGenerator(
          (rate): RandUniformTensorGenerator(low=1.0, high=10.0)
          (max_value): RandUniformTensorGenerator(low=1.0, high=100.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self, rate: BaseTensorGenerator | dict, max_value: BaseTensorGenerator | dict
    ) -> None:
        super().__init__()
        self._rate = setup_tensor_generator(rate)
        self._max_value = setup_tensor_generator(max_value)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"rate": self._rate, "max_value": self._max_value}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return trunc_exponential(
            rate=self._rate.generate(size=size, rng=rng),
            max_value=self._max_value.generate(size=size, rng=rng),
            generator=rng,
        )



from __future__ import annotations

__all__ = ["BaseTensorGenerator", "is_tensor_generator_config", "setup_tensor_generator"]

import logging
from abc import ABC, abstractmethod

from objectory import AbstractFactory
from objectory.utils import is_object_config
from torch import Generator, Tensor

from startorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


class BaseTensorGenerator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to generate tensor.

    A child class has to implement the ``generate`` method.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.tensor import RandUniform
        >>> generator = RandUniform()
        >>> generator
        RandUniformTensorGenerator(low=0.0, high=1.0)
        >>> generator.generate(size=(4, 12))
        tensor([[...]])
    """

    @abstractmethod
    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        r"""Generates a tensor.

        Args:
        ----
            size (tuple): Specifies the size of the tensor to generate.
            rng (``torch.Generator`` or None, optional): Specifies
                an optional random number generator. Default: ``None``

        Returns:
        -------
            ``torch.Tensor``: The generated tensor with the specified
                size.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from startorch.tensor import RandUniform
            >>> generator = RandUniform()
            >>> generator.generate(size=(4, 12))
            tensor([[...]])
        """


def is_tensor_generator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseTensorGenerator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseTensorGenerator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import is_tensor_generator_config
        >>> is_tensor_generator_config({"_target_": "startorch.tensor.RandUniform"})
        True
    """
    return is_object_config(config, BaseTensorGenerator)


def setup_tensor_generator(generator: BaseTensorGenerator | dict) -> BaseTensorGenerator:
    r"""Sets up a tensor generator.

    The tensor generator is instantiated from its configuration by
    using the ``BaseTensorGenerator`` factory function.

    Args:
    ----
        generator (``BaseTensorGenerator`` or dict): Specifies a
            tensor generator or its configuration.

    Returns:
    -------
        ``BaseTensorGenerator``: A tensor generator.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import setup_tensor_generator
        >>> setup_tensor_generator({"_target_": "startorch.tensor.RandUniform"})
        RandUniformTensorGenerator(low=0.0, high=1.0)
    """
    if isinstance(generator, dict):
        logger.info(
            "Initializing a tensor generator from its configuration... "
            f"{str_target_object(generator)}"
        )
        generator = BaseTensorGenerator.factory(**generator)
    if not isinstance(generator, BaseTensorGenerator):
        logger.warning(f"generator is not a `BaseTensorGenerator` (received: {type(generator)})")
    return generator



from __future__ import annotations

__all__ = [
    "AsinhUniformTensorGenerator",
    "LogUniformTensorGenerator",
    "RandAsinhUniformTensorGenerator",
    "RandIntTensorGenerator",
    "RandLogUniformTensorGenerator",
    "RandUniformTensorGenerator",
    "UniformTensorGenerator",
]


from coola.utils.format import str_indent, str_mapping
from torch import Generator, Tensor, randint

from startorch.random import (
    asinh_uniform,
    log_uniform,
    rand_asinh_uniform,
    rand_log_uniform,
    rand_uniform,
    uniform,
)
from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator


class AsinhUniformTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator to generate tensors by sampling
    values from an asinh-uniform distribution.

    Args:
    ----
        low (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the minimum
            value (inclusive).
        high (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the maximum
            value (exclusive).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform, AsinhUniform
        >>> generator = AsinhUniform(
        ...     low=RandUniform(low=-1000, high=-100), high=RandUniform(low=100, high=1000)
        ... )
        >>> generator
        AsinhUniformTensorGenerator(
          (low): RandUniformTensorGenerator(low=-1000.0, high=-100.0)
          (high): RandUniformTensorGenerator(low=100.0, high=1000.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        low: BaseTensorGenerator | dict,
        high: BaseTensorGenerator | dict,
    ) -> None:
        super().__init__()
        self._low = setup_tensor_generator(low)
        self._high = setup_tensor_generator(high)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"low": self._low, "high": self._high}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return asinh_uniform(
            low=self._low.generate(size=size, rng=rng),
            high=self._high.generate(size=size, rng=rng),
            generator=rng,
        )


class LogUniformTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator to generate tensors by sampling
    values from a log-uniform distribution.

    Args:
    ----
        low (``BaseSequenceGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the minimum
            value (inclusive).
        high (``BaseSequenceGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the maximum
            value (exclusive).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform, LogUniform
        >>> generator = LogUniform(
        ...     low=RandUniform(low=0.1, high=1.0), high=RandUniform(low=100, high=1000)
        ... )
        >>> generator
        LogUniformTensorGenerator(
          (low): RandUniformTensorGenerator(low=0.1, high=1.0)
          (high): RandUniformTensorGenerator(low=100.0, high=1000.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        low: BaseTensorGenerator | dict,
        high: BaseTensorGenerator | dict,
    ) -> None:
        super().__init__()
        self._low = setup_tensor_generator(low)
        self._high = setup_tensor_generator(high)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"low": self._low, "high": self._high}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return log_uniform(
            low=self._low.generate(size=size, rng=rng),
            high=self._high.generate(size=size, rng=rng),
            generator=rng,
        )


class RandAsinhUniformTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator by sampling values from an asinh-
    uniform distribution.

    Args:
    ----
        low (float): Specifies the minimum value (inclusive).
        high (float): Specifies the maximum value (exclusive).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandAsinhUniform
        >>> generator = RandAsinhUniform(low=-1000, high=1000)
        >>> generator
        RandAsinhUniformTensorGenerator(low=-1000.0, high=1000.0)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, low: float, high: float) -> None:
        super().__init__()
        self._low = float(low)
        if high < low:
            raise ValueError(f"high ({high}) has to be greater or equal to low ({low})")
        self._high = float(high)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(low={self._low}, " f"high={self._high})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_asinh_uniform(
            size=size,
            low=self._low,
            high=self._high,
            generator=rng,
        )


class RandIntTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator by sampling integer values from a
    uniform distribution.

    Args:
    ----
        low (int, optional): Specifies the minimum value
            (inclusive).
        high (int, optional): Specifies the maximum value
            (exclusive).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandInt
        >>> generator = RandInt(low=0, high=10)
        >>> generator
        RandIntTensorGenerator(low=0, high=10)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, low: int, high: int) -> None:
        super().__init__()
        self._low = int(low)
        if high <= low:
            raise ValueError(f"high ({high}) has to be greater than low ({low})")
        self._high = int(high)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(low={self._low}, high={self._high})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return randint(size=size, low=self._low, high=self._high, generator=rng)


class RandLogUniformTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator to generate tensors by sampling
    values from a log-uniform distribution.

    Args:
    ----
        low (float): Specifies the minimum value (inclusive).
        high (float): Specifies the maximum value (exclusive).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandLogUniform
        >>> generator = RandLogUniform(low=0.1, high=100.0)
        >>> generator
        RandLogUniformTensorGenerator(low=0.1, high=100.0)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, low: float, high: float) -> None:
        super().__init__()
        self._low = float(low)
        if high < low:
            raise ValueError(f"high ({high}) has to be greater or equal to low ({low})")
        self._high = float(high)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(low={self._low}, high={self._high})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_log_uniform(
            size=size,
            low=self._low,
            high=self._high,
            generator=rng,
        )


class RandUniformTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator by sampling values from a uniform
    distribution.

    Args:
    ----
        low (float, optional): Specifies the minimum value
            (inclusive). Default: ``0.0``
        high (float, optional): Specifies the maximum value
            (exclusive). Default: ``1.0``

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform
        >>> generator = RandUniform(low=0, high=10)
        >>> generator
        RandUniformTensorGenerator(low=0.0, high=10.0)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, low: float = 0.0, high: float = 1.0) -> None:
        super().__init__()
        self._low = float(low)
        if high < low:
            raise ValueError(f"high ({high}) has to be greater or equal to low ({low})")
        self._high = float(high)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(low={self._low}, high={self._high})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_uniform(size=size, low=self._low, high=self._high, generator=rng)


class UniformTensorGenerator(BaseTensorGenerator):
    r"""Implements a tensor generator by sampling values from a uniform
    distribution.

    Args:
    ----
        low (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the minimum
            value (inclusive).
        high (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the maximum
            value (exclusive).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform, Uniform
        >>> generator = UniformTensorGenerator(
        ...     low=RandUniform(low=0, high=2), high=RandUniform(low=8, high=10)
        ... )
        >>> generator
        UniformTensorGenerator(
          (low): RandUniformTensorGenerator(low=0.0, high=2.0)
          (high): RandUniformTensorGenerator(low=8.0, high=10.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, low: BaseTensorGenerator | dict, high: BaseTensorGenerator | dict) -> None:
        super().__init__()
        self._low = setup_tensor_generator(low)
        self._high = setup_tensor_generator(high)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"low": self._low, "high": self._high}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return uniform(
            low=self._low.generate(size, rng=rng),
            high=self._high.generate(size, rng=rng),
            generator=rng,
        )



from __future__ import annotations

__all__ = [
    "AcoshSequenceGenerator",
    "AsinhSequenceGenerator",
    "AtanhSequenceGenerator",
    "CoshSequenceGenerator",
    "SinhSequenceGenerator",
    "TanhSequenceGenerator",
]

from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.wrapper import BaseWrapperSequenceGenerator


class AcoshSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that computes the inverse
    hyperbolic cosine (arccosh) of each value.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Acosh, RandUniform
        >>> generator = Acosh(RandUniform())
        >>> generator
        AcoshSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).acosh()


class AsinhSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that computes the inverse
    hyperbolic sine (arcsinh) of each value.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Asinh, RandUniform
        >>> generator = Asinh(RandUniform())
        >>> generator
        AsinhSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).asinh()


class AtanhSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that computes the inverse
    hyperbolic tangent (arctanh) of each value.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Atanh, RandUniform
        >>> generator = Atanh(RandUniform())
        >>> generator
        AtanhSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).atanh()


class CoshSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that computes the hyperbolic
    cosine (cosh) of each value.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Cosh, RandUniform
        >>> generator = Cosh(RandUniform())
        >>> generator
        CoshSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).cosh()


class SinhSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that computes the hyperbolic sine
    (sinh) of each value.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Sinh, RandUniform
        >>> generator = Sinh(RandUniform())
        >>> generator
        SinhSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).sinh()


class TanhSequenceGenerator(BaseWrapperSequenceGenerator):
    r"""Implements a sequence generator that computes the hyperbolic
    tangent (tanh) of each value.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.sequence import Tanh, RandUniform
        >>> generator = Tanh(RandUniform())
        >>> generator
        TanhSequenceGenerator(
          (sequence): RandUniformSequenceGenerator(low=0.0, high=1.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=12, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return self._generator.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).tanh()



from __future__ import annotations

__all__ = [
    "HalfCauchyTensorGenerator",
    "RandHalfCauchyTensorGenerator",
    "RandTruncHalfCauchyTensorGenerator",
    "TruncHalfCauchyTensorGenerator",
]


from coola.utils.format import str_indent, str_mapping
from torch import Generator, Tensor

from startorch.random import (
    half_cauchy,
    rand_half_cauchy,
    rand_trunc_half_cauchy,
    trunc_half_cauchy,
)
from startorch.tensor.base import BaseTensorGenerator, setup_tensor_generator


class HalfCauchyTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensor by sampling values from a
    half-Cauchy distribution.

    Args:
    ----
        scale (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the scale.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import HalfCauchy, RandUniform
        >>> generator = HalfCauchy(scale=RandUniform(low=1.0, high=2.0))
        >>> generator
        HalfCauchyTensorGenerator(
          (scale): RandUniformTensorGenerator(low=1.0, high=2.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, scale: BaseTensorGenerator | dict) -> None:
        super().__init__()
        self._scale = setup_tensor_generator(scale)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"scale": self._scale}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return half_cauchy(
            scale=self._scale.generate(size=size, rng=rng),
            generator=rng,
        )


class RandHalfCauchyTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensor by sampling values from a
    half-Cauchy distribution.

    Args:
    ----
        scale (float, optional): Specifies the scale of the
            distribution. Default: ``1.0``

    Raises:
    ------
        ValueError if ``scale`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandHalfCauchy
        >>> generator = RandHalfCauchy(scale=1.0)
        >>> generator
        RandHalfCauchyTensorGenerator(scale=1.0)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(self, scale: float = 1.0) -> None:
        super().__init__()
        if scale <= 0:
            raise ValueError(f"scale has to be greater than 0 (received: {scale})")
        self._scale = float(scale)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(scale={self._scale})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_half_cauchy(
            size=size,
            scale=self._scale,
            generator=rng,
        )


class RandTruncHalfCauchyTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensor by sampling values from a
    truncated half-Cauchy distribution.

    Args:
    ----
        scale (float, optional): Specifies the scale of the
            distribution. Default: ``1.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``4.0``

    Raises:
    ------
        ValueError if ``scale`` is not a positive number.
        ValueError if ``max_value`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandTruncHalfCauchy
        >>> generator = RandTruncHalfCauchy(scale=1.0, max_value=5.0)
        >>> generator
        RandTruncHalfCauchyTensorGenerator(scale=1.0, max_value=5.0)
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        scale: float = 1.0,
        max_value: float = 4.0,
    ) -> None:
        super().__init__()
        if scale <= 0:
            raise ValueError(f"scale has to be greater than 0 (received: {scale})")
        self._scale = float(scale)
        if max_value <= 0:
            raise ValueError(f"max_value has to be greater than 0 (received: {max_value})")
        self._max_value = float(max_value)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(scale={self._scale}, max_value={self._max_value})"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return rand_trunc_half_cauchy(
            size=size,
            scale=self._scale,
            max_value=self._max_value,
            generator=rng,
        )


class TruncHalfCauchyTensorGenerator(BaseTensorGenerator):
    r"""Implements a class to generate tensor by sampling values from a
    half-Cauchy distribution.

    Args:
    ----
        scale (``BaseTensorGenerator`` or dict): Specifies a tensor
            generator (or its configuration) to generate the scale.
        max_value (``BaseTensorGenerator`` or dict): Specifies a
            tensor generator (or its configuration) to generate the
            maximum value (excluded).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform, TruncHalfCauchy
        >>> generator = TruncHalfCauchy(
        ...     scale=RandUniform(low=1.0, high=2.0),
        ...     max_value=RandUniform(low=5.0, high=10.0),
        ... )
        >>> generator
        TruncHalfCauchyTensorGenerator(
          (scale): RandUniformTensorGenerator(low=1.0, high=2.0)
          (max_value): RandUniformTensorGenerator(low=5.0, high=10.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def __init__(
        self,
        scale: BaseTensorGenerator | dict,
        max_value: BaseTensorGenerator | dict,
    ) -> None:
        super().__init__()
        self._scale = setup_tensor_generator(scale)
        self._max_value = setup_tensor_generator(max_value)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"scale": self._scale, "max_value": self._max_value}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return trunc_half_cauchy(
            scale=self._scale.generate(size=size, rng=rng),
            max_value=self._max_value.generate(size=size, rng=rng),
            generator=rng,
        )



from __future__ import annotations

__all__ = ["prepare_probabilities", "prepare_weighted_generators"]

from collections.abc import Sequence
from typing import Any

import torch
from torch import Tensor

GENERATOR = "generator"
WEIGHT = "weight"


def prepare_probabilities(weights: Tensor | Sequence[int | float]) -> torch.Tensor:
    r"""Converts un-normalized positive weights to probabilities.

    Args:
    ----
        weights (``torch.Tensor`` of shape ``(num_categories,)`` and
            type float or ``Sequence``): Specifies the vector of
            weights associated to each category. The weights have
            to be positive.

    Returns:
    -------
        ``torch.Tensor`` of type float and shape ``(num_categories,)``:
            The vector of probability associated at each category.

    Raises:
    ------
        ValueError if the weights are not valid.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.utils.weight import prepare_probabilities
        >>> prepare_probabilities([1, 1, 1, 1])
        tensor([0.2500, 0.2500, 0.2500, 0.2500])
    """
    if not torch.is_tensor(weights):
        weights = torch.as_tensor(weights)
    if weights.ndim != 1:
        raise ValueError(f"weights has to be a 1D tensor (received a {weights.ndim}D tensor)")
    if weights.min() < 0:
        raise ValueError(
            f"The values in weights have to be positive (min: {weights.min()}  weights: {weights})"
        )
    if weights.sum() == 0:
        raise ValueError(
            f"The sum of the weights has to be greater than 0 (sum: {weights.sum()}  "
            f"weights: {weights})"
        )
    return weights.float() / weights.sum()


def prepare_weighted_generators(
    generators: Sequence[dict],
) -> tuple[tuple[Any, ...], tuple[float, ...]]:
    r"""Prepare the tensor generators.

    Each dictionary in the input tuple/list should have the
    following items:

        - a key ``'generator'`` which indicates the tensor generator
            or its configuration.
        - an optional key ``'weight'`` with a float value which
            indicates the weight of the tensor generator.
            If this key is absent, the weight is set to ``1.0``.

    Args:
    ----
        generators (tuple or list): Specifies the tensor generators
            and their weights. See above to learn about the
            expected format.

    Returns:
    -------
        tuple with two values:
            - a tuple of generators or their configurations
            - a tuple of floats

    Example usage:

    .. code-block:: pycon

        >>> from startorch.utils.weight import prepare_weighted_generators
        >>> from startorch.tensor import RandUniform, RandNormal
        >>> prepare_weighted_generators(
        ...     (
        ...         {"weight": 2.0, "generator": RandUniform()},
        ...         {"weight": 1.0, "generator": RandNormal()},
        ...     )
        ... )
        ((RandUniformTensorGenerator(low=0.0, high=1.0), RandNormalTensorGenerator(mean=0.0, std=1.0)), (2.0, 1.0))
    """
    gens = []
    weights = []
    for generator in generators:
        gens.append(generator[GENERATOR])
        weights.append(float(generator.get(WEIGHT, 1.0)))
    return tuple(gens), tuple(weights)



r"""This module implements functions to sample values from continuous
univariate distribution supported on a bounded interval."""

from __future__ import annotations

__all__ = [
    "asinh_uniform",
    "log_uniform",
    "rand_asinh_uniform",
    "rand_log_uniform",
    "rand_trunc_cauchy",
    "rand_trunc_exponential",
    "rand_trunc_half_cauchy",
    "rand_trunc_half_normal",
    "rand_trunc_log_normal",
    "rand_trunc_normal",
    "rand_uniform",
    "trunc_cauchy",
    "trunc_exponential",
    "trunc_half_cauchy",
    "trunc_half_normal",
    "trunc_log_normal",
    "trunc_normal",
    "uniform",
]

import math

import torch
from torch import Tensor
from torch.distributions import (
    Cauchy,
    Exponential,
    HalfCauchy,
    HalfNormal,
    LogNormal,
    Normal,
)

from startorch.utils.tensor import shapes_are_equal


def rand_trunc_cauchy(
    size: list[int] | tuple[int, ...],
    loc: float = 0.0,
    scale: float = 1.0,
    min_value: float = -2.0,
    max_value: float = 2.0,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a truncated
    Cauchy distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        loc (float, optional): Specifies the location of the Cauchy
            distribution. Default: ``0.0``
        scale (float, optional): Specifies the scale of the Cauchy
            distribution. Default: ``1.0``
        min_value (float, optional): Specifies the minimum value.
            Default: ``-2.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``2.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from
            a truncated Cauchy distribution

    Raises:
    ------
        ValueError if the ``max_value`` and ``min_value`` parameters
            are not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_trunc_cauchy
        >>> rand_trunc_cauchy((2, 3), loc=1.0, scale=2.0, min_value=-3.0, max_value=3.0)
        tensor([[...]])
    """
    if max_value < min_value:
        raise ValueError(
            f"`max_value` ({max_value}) has to be greater or equal to `min_value` ({min_value})"
        )
    distribution = Cauchy(loc=loc, scale=scale)
    # Values are generated by using a truncated uniform distribution and
    # then using the inverse CDF for the Cauchy distribution.
    # Get upper and lower CDF values
    lo, up = distribution.cdf(torch.tensor([min_value, max_value]))
    return distribution.icdf(rand_uniform(size, low=lo, high=up, generator=generator)).clamp(
        min=min_value, max=max_value
    )


def trunc_cauchy(
    loc: Tensor,
    scale: Tensor,
    min_value: Tensor,
    max_value: Tensor,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a truncated
    Cauchy distribution.

    Args:
    ----
        loc (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the location/median of
            the Cauchy distribution.
        scale (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the scale
            of the Cauchy distribution.
        min_value (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the minimum value.
        max_value (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the maximum value.
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from
            a truncated Cauchy distribution

    Raises:
    ------
        ValueError if the ``loc``, ``scale``, ``max_value`` and
            ``min_value`` parameters are not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import trunc_cauchy
        >>> trunc_cauchy(
        ...     loc=torch.tensor([1.0, 0.0, -1.0]),
        ...     scale=torch.tensor([1.0, 3.0, 5.0]),
        ...     min_value=torch.tensor([-5.0, -10.0, -15.0]),
        ...     max_value=torch.tensor([5.0, 10.0, 15.0]),
        ... )
        tensor([...])
    """
    if not shapes_are_equal((loc, scale, min_value, max_value)):
        raise ValueError(
            f"Incorrect shapes. The shapes of all the input tensors must be equal: loc={loc.shape}"
            f"  scale={scale.shape}  min_value={min_value.shape}  max_value={max_value.shape}"
        )
    if torch.any(scale <= 0.0):
        raise ValueError("All the `scale` values have to be greater than 0")
    if torch.any(max_value < min_value):
        raise ValueError(
            "Found at least one value in `min_value` that is higher than its associated "
            "`max_value`"
        )
    distribution = Cauchy(loc=loc, scale=scale)
    # Values are generated by using a truncated uniform distribution and
    # then using the inverse CDF for the Cauchy distribution.
    # Get upper and lower CDF values
    lo, up = distribution.cdf(torch.stack([min_value, max_value]))
    return distribution.icdf(uniform(low=lo, high=up, generator=generator)).clamp(
        min=min_value, max=max_value
    )


def rand_trunc_exponential(
    size: list[int] | tuple[int, ...],
    rate: float = 1.0,
    max_value: float = 5.0,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a truncated
    Exponential distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        rate (float, optional): Specifies the rate of the Exponential
            distribution. Default: ``1.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``5.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            truncated Exponential distribution

    Raises:
    ------
        ValueError if the ``max_value`` parameter is not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_trunc_exponential
        >>> rand_trunc_exponential((2, 3), rate=1.0, max_value=3.0)
        tensor([[...]])
    """
    if max_value <= 0:
        raise ValueError(f"`max_value` has to be greater than 0 (received: {max_value})")
    distribution = Exponential(rate=rate)
    # Values are generated by using a truncated uniform distribution and
    # then using the inverse CDF for the Exponential distribution.
    # Get upper and lower CDF values
    up = distribution.cdf(torch.tensor(max_value)).item()
    return distribution.icdf(rand_uniform(size, low=0.0, high=up, generator=generator)).clamp(
        max=max_value
    )


def trunc_exponential(
    rate: Tensor,
    max_value: Tensor,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a truncated
    Exponential distribution.

    Args:
    ----
        rate (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the rate
            of the Exponential distribution.
        max_value (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the maximum value.
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from
            a truncated Exponential distribution.

    Raises:
    ------
        ValueError if the ``rate`` and ``max_value`` parameter are not
            valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import trunc_exponential
        >>> trunc_exponential(
        ...     rate=torch.tensor([1.0, 3.0, 5.0]),
        ...     max_value=torch.tensor([5.0, 10.0, 15.0]),
        ... )
        tensor([...])
    """
    if rate.shape != max_value.shape:
        raise ValueError(
            "Incorrect shapes. The shapes of all the input tensors must be equal:  "
            f"rate={rate.shape}  max_value={max_value.shape})"
        )
    if torch.any(rate <= 0.0):
        raise ValueError("All the `rate` values have to be greater than 0")
    if torch.any(max_value <= 0.0):
        raise ValueError("Found at least one value in `max_value` that is lower or equal to 0")
    distribution = Exponential(rate=rate)
    # Values are generated by using a truncated uniform distribution and
    # then using the inverse CDF for the Exponential distribution.
    # Get upper and lower CDF values
    u = distribution.cdf(max_value)
    return distribution.icdf(torch.rand(rate.shape, generator=generator).mul(u)).clamp(
        max=max_value
    )


def rand_trunc_half_cauchy(
    size: list[int] | tuple[int, ...],
    scale: float = 1.0,
    max_value: float = 4.0,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a truncated
    half-Cauchy distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        scale (float, optional): Specifies the scale of the
            half-Cauchy distribution. Default: ``1.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``4.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            truncated half-Cauchy distribution.

    Raises:
    ------
        ValueError if the ``max_value`` parameter is not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_trunc_half_cauchy
        >>> rand_trunc_half_cauchy((2, 3), scale=1.0, max_value=3.0)
        tensor([[...]])
    """
    if max_value <= 0:
        raise ValueError(f"`max_value` has to be greater than 0 (received: {max_value})")
    distribution = HalfCauchy(scale=scale)
    # Values are generated by using a truncated uniform distribution and
    # then using the inverse CDF for the Cauchy distribution.
    # Get upper CDF value
    u = distribution.cdf(torch.tensor(max_value))
    return distribution.icdf(rand_uniform(size, low=0.0, high=u, generator=generator)).clamp(
        max=max_value
    )


def trunc_half_cauchy(
    scale: Tensor,
    max_value: Tensor,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a truncated
    half-Cauchy distribution.

    Args:
    ----
        scale (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the scale
            of the half-Cauchy distribution.
        max_value (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the maximum value.
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from
            a truncated Cauchy distribution

    Raises:
    ------
        ValueError if the ``scale`` and ``max_value`` parameter are
            not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import trunc_half_cauchy
        >>> trunc_half_cauchy(
        ...     scale=torch.tensor([1.0, 3.0, 5.0]),
        ...     max_value=torch.tensor([5.0, 10.0, 15.0]),
        ... )
        tensor([...])
    """
    if scale.shape != max_value.shape:
        raise ValueError(
            "Incorrect shapes. The shapes of all the input tensors must be equal:  "
            f"scale={scale.shape}  max_value={max_value.shape}"
        )
    if torch.any(scale <= 0.0):
        raise ValueError("All the `scale` values have to be greater than 0")
    if torch.any(max_value <= 0.0):
        raise ValueError("Found at least one value in `max_value` that is lower or equal to 0")
    distribution = HalfCauchy(scale=scale)
    # Values are generated by using a truncated uniform distribution and
    # then using the inverse CDF for the Cauchy distribution.
    # Get upper CDF value
    u = distribution.cdf(max_value)
    return distribution.icdf(torch.rand(scale.shape, generator=generator).mul(u)).clamp(
        max=max_value
    )


def rand_trunc_half_normal(
    size: list[int] | tuple[int, ...],
    std: float = 1.0,
    max_value: float = 5.0,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a truncated
    half-Normal distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        std (float, optional): Specifies the standard deviation of
            the half-Normal distribution. Default: ``1.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``5.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            truncated half-Normal distribution.

    Raises:
    ------
        ValueError if the ``max_value`` parameter is not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_trunc_half_normal
        >>> rand_trunc_half_normal((2, 3), std=1.0, max_value=3.0)
        tensor([[...]])
    """
    if max_value <= 0:
        raise ValueError(f"`max_value` has to be greater than 0 (received: {max_value})")
    distribution = HalfNormal(scale=std)
    # Values are generated by using a truncated uniform distribution and
    # then using the inverse CDF for the Normal distribution.
    # Get upper cdf value
    u = distribution.cdf(torch.tensor(max_value))
    return distribution.icdf(rand_uniform(size, low=0.0, high=u, generator=generator)).clamp(
        max=max_value
    )


def trunc_half_normal(
    std: Tensor,
    max_value: Tensor,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a truncated
    half-Normal distribution.

    Args:
    ----
        std (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the standard deviation
            of the half-Normal distribution.
        max_value (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the maximum value.
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            truncated half-Normal distribution.

    Raises:
    ------
        ValueError if the ``std`` and ``max_value`` parameters are not
            valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import trunc_half_normal
        >>> trunc_half_normal(
        ...     std=torch.tensor([1.0, 3.0, 5.0]), max_value=torch.tensor([5.0, 10.0, 15.0])
        ... )
        tensor([...])
    """
    if std.shape != max_value.shape:
        raise ValueError(
            "Incorrect shapes. The shapes of all the input tensors must be equal: "
            f"std={std.shape}  max_value={max_value.shape}"
        )
    if torch.any(std <= 0.0):
        raise ValueError("All the `std` values have to be greater than 0")
    if torch.any(max_value <= 0.0):
        raise ValueError("All the `max_value` values must be greater than 0")
    distribution = HalfNormal(scale=std)
    # Values are generated by using a uniform distribution and
    # then using the inverse CDF for the Normal distribution.
    # Get upper and lower CDF values
    u = distribution.cdf(max_value)
    return distribution.icdf(torch.rand(std.shape, generator=generator).mul(u)).clamp(max=max_value)


def rand_trunc_log_normal(
    size: list[int] | tuple[int, ...],
    mean: float = 0.0,
    std: float = 1.0,
    min_value: float = 0.0,
    max_value: float = 5.0,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a truncated log-
    Normal distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        mean (float, optional): Specifies the mean of the underlying
            Normal distribution. Default: ``0.0``
        std (float, optional): Specifies the standard deviation of
            the underlying Normal distribution. Default: ``1.0``
        min_value (float, optional): Specifies the minimum value.
            Default: ``0.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``5.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            truncated log-Normal distribution.

    Raises:
    ------
        ValueError if the ``min_value`` and ``max_value`` parameters
            are not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_trunc_log_normal
        >>> rand_trunc_log_normal((2, 3), mean=0.0, std=1.0, min_value=1.0, max_value=4.0)
        tensor([[...]])
    """
    if max_value < min_value:
        raise ValueError(
            f"`max_value` ({max_value}) has to be greater or equal to `min_value` ({min_value})"
        )
    distribution = LogNormal(loc=mean, scale=std)
    # Values are generated by using a truncated uniform distribution and
    # then using the inverse CDF for the log-Normal distribution.
    # Get upper and lower CDF values
    lo, up = distribution.cdf(torch.tensor([min_value, max_value]))
    return distribution.icdf(rand_uniform(size, low=lo, high=up, generator=generator)).clamp(
        min=min_value, max=max_value
    )


def trunc_log_normal(
    mean: Tensor,
    std: Tensor,
    min_value: Tensor,
    max_value: Tensor,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a truncated log-
    Normal distribution.

    Args:
    ----
        mean (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the mean of the
            underlying Normal distribution.
        std (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the standard deviation
            of the underlying Normal distribution.
        min_value (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the minimum value.
        max_value (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the maximum value.
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            truncated log-Normal distribution.

    Raises:
    ------
        ValueError if the ``mean``, ``std``, ``min_value`` and
            ``max_value`` parameters are not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import trunc_log_normal
        >>> trunc_log_normal(
        ...     mean=torch.tensor([-1.0, 0.0, 1.0]),
        ...     std=torch.tensor([1.0, 3.0, 5.0]),
        ...     min_value=torch.tensor([0.0, 1.0, 2.0]),
        ...     max_value=torch.tensor([5.0, 10.0, 15.0]),
        ... )
        tensor([...])
    """
    if not shapes_are_equal((mean, std, min_value, max_value)):
        raise ValueError(
            "Incorrect shapes. The shapes of all the input tensors must be equal: "
            f"mean={mean.shape}  std={std.shape}  min_value={min_value.shape}  "
            f"max_value={max_value.shape}"
        )
    if torch.any(std <= 0.0):
        raise ValueError("All the `std` values have to be greater than 0")
    if torch.any(max_value < min_value):
        raise ValueError(
            "Found at least one value in `min_value` that is higher than its associated "
            "`max_value`"
        )
    distribution = LogNormal(loc=mean, scale=std)
    # Values are generated by using a uniform distribution and
    # then using the inverse CDF for the log-Normal distribution.
    # Get upper and lower CDF values
    lo, up = distribution.cdf(torch.stack([min_value, max_value]))
    return distribution.icdf(uniform(low=lo, high=up, generator=generator)).clamp(
        min=min_value, max=max_value
    )


def rand_trunc_normal(
    size: list[int] | tuple[int, ...],
    mean: float = 0.0,
    std: float = 1.0,
    min_value: float = -3.0,
    max_value: float = 3.0,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a truncated
    Normal distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        mean (float, optional): Specifies the mean of the Normal
            distribution. Default: ``0.0``
        std (float, optional): Specifies the standard deviation of
            the Normal distribution. Default: ``1.0``
        min_value (float, optional): Specifies the minimum value.
            Default: ``-3.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``3.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            truncated Normal distribution.

    Raises:
    ------
        ValueError if the ``min_value`` and  ``max_value`` parameters
            are not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_trunc_normal
        >>> rand_trunc_normal((2, 3), mean=1.0, std=2.0, min_value=-5.0, max_value=5.0)
        tensor([[...]])
    """
    if max_value < min_value:
        raise ValueError(
            f"`max_value` ({max_value}) has to be greater or equal to `min_value` ({min_value})"
        )
    distribution = Normal(loc=mean, scale=std)
    # Values are generated by using a truncated uniform distribution and
    # then using the inverse CDF for the Normal distribution.
    # Get upper and lower cdf values
    lo, up = distribution.cdf(torch.tensor([min_value, max_value]))
    return distribution.icdf(rand_uniform(size, low=lo, high=up, generator=generator)).clamp(
        min=min_value, max=max_value
    )


def trunc_normal(
    mean: Tensor,
    std: Tensor,
    min_value: Tensor,
    max_value: Tensor,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a truncated
    Normal distribution.

    Args:
    ----
        mean (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the mean of the
            Normal distribution.
        std (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the standard deviation
            of the Normal distribution.
        min_value (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the minimum value.
        max_value (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the maximum value.
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            truncated Normal distribution.

    Raises:
    ------
        ValueError if the ``min_value`` and  ``max_value`` parameters
            are not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import trunc_normal
        >>> trunc_normal(
        ...     mean=torch.tensor([1.0, 0.0, -1.0]),
        ...     std=torch.tensor([1.0, 3.0, 5.0]),
        ...     min_value=torch.tensor([-5.0, -10.0, -15.0]),
        ...     max_value=torch.tensor([5.0, 10.0, 15.0]),
        ... )
        tensor([...])
    """
    if not shapes_are_equal((mean, std, min_value, max_value)):
        raise ValueError(
            "Incorrect shapes. The shapes of all the input tensors must be equal: "
            f"mean={mean.shape}  std={std.shape}  min_value={min_value.shape}  "
            f"max_value={max_value.shape}"
        )
    if torch.any(std <= 0.0):
        raise ValueError("All the `std` values have to be greater than 0")
    if torch.any(max_value < min_value):
        raise ValueError(
            "Found at least one value in `min_value` that is higher than its associated "
            "`max_value`"
        )
    distribution = Normal(loc=mean, scale=std)
    # Values are generated by using a uniform distribution and
    # then using the inverse CDF for the Normal distribution.
    # Get upper and lower CDF values
    lo, up = distribution.cdf(torch.stack([min_value, max_value]))
    return distribution.icdf(uniform(low=lo, high=up, generator=generator)).clamp(
        min=min_value, max=max_value
    )


def rand_uniform(
    size: list[int] | tuple[int, ...],
    low: float = 0.0,
    high: float = 1.0,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a uniform
    distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        low (float, optional): Specifies the minimum value
            (inclusive). Default: ``0.0``
        high (float, optional): Specifies the maximum value
            (exclusive). Default: ``1.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            uniform distribution.

    Raises:
    ------
        ValueError if the ``low`` and  ``high`` parameters  are not
            valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_uniform
        >>> rand_uniform((2, 3), low=-1.0, high=2.0)
        tensor([[...]])
    """
    if high < low:
        raise ValueError(f"`high` ({high}) has to be greater or equal to `low` ({low})")
    return (
        torch.rand(size=size, generator=generator).mul(high - low).add(low).clamp(min=low, max=high)
    )


def uniform(low: Tensor, high: Tensor, generator: torch.Generator | None = None) -> Tensor:
    r"""Creates a tensor filled with values sampled from a uniform
    distribution.

    Unlike ``rand_uniform``, this function allows to sample values
    from different uniform distributions at the same time.
    The shape of the ``low`` and ``high`` tensors are used to infer
    the output size.

    Args:
    ----
        low (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the minimum values
            (inclusive).
        high (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the maximum values
            (exclusive).
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``: A tensor filled with values sampled
            from a uniform distribution where the minimum and maximum
            values are given as input.

    Raises:
        ValueError if the input tensor shapes do not match or if at
            least one value in ``low`` tensor is higher than its
            associated high value in ``high``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import uniform
        >>> uniform(low=torch.tensor([-1.0, 0.0, 1.0]), high=torch.tensor([1.0, 3.0, 5.0]))
        tensor([...])
    """
    if low.shape != high.shape:
        raise ValueError(
            f"Incorrect shapes. The shapes of all the input tensors must be equal: "
            f"low={low.shape}  high={high.shape})"
        )
    if torch.any(high < low):
        raise ValueError(
            "Found at least one value in `low` that is higher than its associated "
            "value in `high`.`uniform` expects to return a [low, high) range"
        )
    return (
        torch.rand(size=high.shape, generator=generator)
        .mul(high - low)
        .add(low)
        .clamp(min=low, max=high)
    )


def rand_log_uniform(
    size: list[int] | tuple[int, ...],
    low: float,
    high: float,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a uniform
    distribution in the log space.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        low (float): Specifies the minimum value (inclusive).
            This value needs to be positive.
        high (float): Specifies the maximum value (exclusive).
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            uniform distribution in the log space.

    Raises:
    ------
        ValueError if the ``low`` and  ``high`` parameters  are not
            valid.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.random import rand_log_uniform
        >>> rand_log_uniform((2, 3), low=0.1, high=1000.0)
        tensor([[...]])
    """
    if high < low:
        raise ValueError(f"`high` ({high}) has to be greater or equal to `low` ({low})")
    log_low = math.log(low)
    return (
        torch.rand(size=size, generator=generator)
        .mul(math.log(high) - log_low)
        .add(log_low)
        .exp()
        .clamp(min=low, max=high)
    )


def log_uniform(low: Tensor, high: Tensor, generator: torch.Generator | None = None) -> Tensor:
    r"""Creates a tensor filled with values sampled from a uniform
    distribution in the log space.

    Args:
    ----
        low (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the minimum values
            (inclusive).
        high (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the maximum values
            (exclusive).
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``: A tensor filled with values sampled
            from a uniform distribution in the log space where the
            minimum and maximum values are given as input.

    Raises:
    ------
        ValueError if the ``low`` and  ``high`` parameters  are not
            valid.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.random import log_uniform
        >>> log_uniform(low=torch.tensor([0.01, 0.1, 1.0]), high=torch.tensor([1.0, 10.0, 100.0]))
        tensor([...])
    """
    if low.shape != high.shape:
        raise ValueError(
            "Incorrect shapes. The shapes of all the input tensors must be equal: "
            f"low={low.shape}  high={high.shape}"
        )
    if torch.any(high < low):
        raise ValueError(
            "Found at least one value in `low` that is higher than its associated "
            "value in `high`. `log_uniform` expects to return a [low, high) range"
        )
    log_low = low.log()
    return (
        torch.rand(size=high.shape, generator=generator)
        .mul(high.log() - log_low)
        .add(log_low)
        .exp()
        .clamp(min=low, max=high)
    )


def rand_asinh_uniform(
    size: list[int] | tuple[int, ...],
    low: float,
    high: float,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a uniform
    distribution in the inverse hyperbolic sine space.

    Args:
        size (list or tuple): Specifies the tensor shape.
        low (float): Specifies the minimum value (inclusive).
            This value needs to be positive.
        high (float): Specifies the maximum value (exclusive).
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
        ``torch.Tensor``: A tensor filled with values sampled from a
            uniform distribution in the inverse hyperbolic sine space.

    Raises:
    ------
        ValueError if the ``low`` and  ``high`` parameters  are not
            valid.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.random import rand_asinh_uniform
        >>> rand_asinh_uniform((2, 3), low=-1000.0, high=1000.0)
        tensor([[...]])
    """
    if high < low:
        raise ValueError(f"`high` ({high}) has to be greater or equal to `low` ({low})")
    log_low = math.asinh(low)
    return (
        torch.rand(size=size, generator=generator)
        .mul(math.asinh(high) - log_low)
        .add(log_low)
        .sinh()
        .clamp(min=low, max=high)
    )


def asinh_uniform(low: Tensor, high: Tensor, generator: torch.Generator | None = None) -> Tensor:
    r"""Creates a tensor filled with values sampled from a uniform
    distribution in the inverse hyperbolic sine space.

    Args:
        low (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the minimum values
            (inclusive).
        high (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the maximum values
            (exclusive).
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
        ``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``: A tensor filled with values sampled
            from a uniform distribution in the inverse hyperbolic sine
            space where the minimum and maximum values are given as
            input.

    Raises:
    ------
        ValueError if the ``low`` and  ``high`` parameters  are not
            valid.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.random import asinh_uniform
        >>> asinh_uniform(
        ...     low=torch.tensor([-10.0, 0.0, 1.0]),
        ...     high=torch.tensor([1.0, 10.0, 100.0]),
        ... )
        tensor([...])
    """
    if low.shape != high.shape:
        raise ValueError(
            "Incorrect shapes. The shapes of all the input tensors must be equal: "
            f"low={low.shape}  high={high.shape}"
        )
    if torch.any(high < low):
        raise ValueError(
            "Found at least one value in `low` that is higher than its associated "
            "value in `high`. `asinh_uniform` expects to return a [low, high) range"
        )
    log_low = low.asinh()
    return (
        torch.rand(size=high.shape, generator=generator)
        .mul(high.asinh() - log_low)
        .add(log_low)
        .sinh()
        .clamp(min=low, max=high)
    )



r"""This module implements functions to sample values from discrete
univariate distribution supported on a bounded interval."""

from __future__ import annotations

__all__ = ["rand_poisson"]


import torch
from torch import Tensor


def rand_poisson(
    size: list[int] | tuple[int, ...],
    rate: float = 1.0,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a Poisson
    distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        rate (float, optional): Specifies the rate of the Poisson
            distribution. This value has to be greater than 0.
            Default: ``1.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor`` of type float: A tensor filled with values
            sampled from a Poisson distribution.

    Raises:
    ------
        ValueError if the ``rate`` parameter is not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_poisson
        >>> rand_poisson(size=(2, 3), rate=2.0)
        tensor([...])
    """
    if rate <= 0:
        raise ValueError(f"rate has to be greater than 0 (received: {rate})")
    return torch.poisson(torch.full(size, rate, dtype=torch.float), generator=generator)



from __future__ import annotations

__all__ = ["cauchy", "normal", "rand_cauchy", "rand_normal"]


import torch
from torch import Tensor


def rand_cauchy(
    size: list[int] | tuple[int, ...],
    loc: float = 0.0,
    scale: float = 1.0,
    generator: torch.Generator | None = None,
) -> torch.Tensor:
    r"""Creates a sequence of continuous variables sampled from a Cauchy
    distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        loc (float, optional): Specifies the location/median of the
            Cauchy distribution. Default: ``0.0``
        scale (float, optional): Specifies the scale of the Cauchy
            distribution. This value has to be greater than 0.
            Default: ``1.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            Cauchy distribution.

    Raises:
    ------
        ValueError if the ``scale`` parameter is not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_cauchy
        >>> rand_cauchy((2, 3), loc=1.0, scale=2.0)
        tensor([[...]])
    """
    if scale <= 0:
        raise ValueError(f"scale has to be greater than 0 (received: {scale})")
    sequence = torch.zeros(*size, dtype=torch.float)
    sequence.cauchy_(median=loc, sigma=scale, generator=generator)
    return sequence


def cauchy(loc: Tensor, scale: Tensor, generator: torch.Generator | None = None) -> Tensor:
    r"""Creates a tensor filled with values sampled from a Cauchy
    distribution.

    Unlike ``rand_cauchy``, this function allows to sample values
    from different Cauchy distributions at the same time.
    The shape of the ``loc`` and ``scale`` tensors are used to infer
    the output size.

    Args:
    ----
        loc (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the location/median
            of the Cauchy distribution.
        scale (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the standard
            deviation of the Cauchy distribution.
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            Cauchy distribution.

    Raises:
    ------
        ValueError if the ``loc`` and ``scale`` parameters are not
            valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import cauchy
        >>> cauchy(loc=torch.tensor([-1.0, 0.0, 1.0]), scale=torch.tensor([1.0, 3.0, 5.0]))
        tensor([...])
    """
    if loc.shape != scale.shape:
        raise ValueError(f"The shapes of loc and scale do not match ({loc.shape} vs {scale.shape})")
    if torch.any(scale <= 0.0):
        raise ValueError(f"scale has to be greater than 0 (received: {scale})")
    return rand_cauchy(loc.shape, generator=generator).mul(scale).add(loc)


def rand_normal(
    size: list[int] | tuple[int, ...],
    mean: float = 0.0,
    std: float = 1.0,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Creates a tensor filled with values sampled from a Normal
    distribution.

    Args:
    ----
        size (list or tuple): Specifies the tensor shape.
        mean (float, optional): Specifies the mean of the Normal
            distribution. Default: ``0.0``
        std (float, optional): Specifies the standard deviation of
            the Normal distribution. Default: ``1.0``
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            Normal distribution.

    Raises:
    ------
        ValueError if the ``std`` parameter is not valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import rand_normal
        >>> rand_normal((2, 3), mean=1.0, std=2.0)
        tensor([[...]])
    """
    if std <= 0.0:
        raise ValueError(f"std has to be greater than 0 (received: {std})")
    return torch.randn(size, generator=generator).mul(std).add(mean)


def normal(mean: Tensor, std: Tensor, generator: torch.Generator | None = None) -> Tensor:
    r"""Creates a tensor filled with values sampled from a Normal
    distribution.

    Args:
    ----
        mean (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the mean.
        std (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the standard
            deviation.
        generator (``torch.Generator`` or None, optional): Specifies
            an optional random generator. Default: ``None``

    Returns:
    -------
        ``torch.Tensor``: A tensor filled with values sampled from a
            Normal distribution.

    Raises:
    ------
        ValueError if the ``mean`` and ``std`` parameters are not
            valid.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.random import normal
        >>> normal(mean=torch.tensor([-1.0, 0.0, 1.0]), std=torch.tensor([1.0, 3.0, 5.0]))
        tensor([...])
    """
    if mean.shape != std.shape:
        raise ValueError(f"The shapes of mean and std do not match ({mean.shape} vs {std.shape})")
    if torch.any(std <= 0.0):
        raise ValueError(f"std has to be greater than 0 (received: {std})")
    return torch.randn(mean.shape, generator=generator).mul(std).add(mean)



from __future__ import annotations

__all__ = ["merge_batches", "scale_batch"]

from collections.abc import Sequence

from redcat import BatchedTensor


def merge_batches(batches: Sequence[BatchedTensor]) -> BatchedTensor:
    r"""Merges batches into a single batch.

    Args:
    ----
        batches (``Sequence``): Specifies the batches to merge.

    Returns:
    -------
        ``BatchedTensor``: The "merged" batch.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from redcat import BatchedTensor
        >>> from startorch.utils.batch import merge_batches
        >>> batch = merge_batches(
        ...     [
        ...         BatchedTensor(torch.arange(6).view(2, 3)),
        ...         BatchedTensor(torch.ones(2, 3)),
        ...         BatchedTensor(torch.zeros(1, 3)),
        ...     ]
        ... )
        >>> batch
        tensor([[0., 1., 2.],
                [3., 4., 5.],
                [1., 1., 1.],
                [1., 1., 1.],
                [0., 0., 0.]], batch_dim=0)
    """
    if not batches:
        raise RuntimeError("No batch is provided")
    batch = batches[0].clone()
    batch.extend(batches[1:])
    return batch


def scale_batch(batch: BatchedTensor, scale: str = "identity") -> BatchedTensor:
    r"""Scales a batch.

    Args:
    ----
        batch (``BatchedTensor``): Specifies the batch to scale.
        scale (str, optional): Specifies the scaling transformation.
            Default: ``'identity'``

    Returns:
    -------
        ``BatchedTensor``: The scaled batch.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from redcat import BatchedTensor
        >>> from startorch.utils.batch import scale_batch
        >>> batch = BatchedTensor(torch.arange(10).view(2, 5))
        >>> scale_batch(batch, scale="asinh")
        tensor([[0.0000, 0.8814, 1.4436, 1.8184, 2.0947],
                [2.3124, 2.4918, 2.6441, 2.7765, 2.8934]], batch_dim=0)
    """
    valid = {"identity", "log", "log10", "log2", "log1p", "asinh"}
    if scale not in valid:
        raise RuntimeError(f"Incorrect scale: {scale}. Valid scales are: {valid}")
    if scale == "identity":
        return batch
    return getattr(batch, scale)()



from __future__ import annotations

__all__ = ["hist_sequence", "plot_sequence"]

from unittest.mock import Mock

from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator
from startorch.utils.batch import merge_batches, scale_batch
from startorch.utils.imports import check_matplotlib, is_matplotlib_available
from startorch.utils.seed import get_torch_generator

if is_matplotlib_available():
    from matplotlib import pyplot as plt
else:
    plt = Mock()  # pragma: no cover


def hist_sequence(
    sequence: BaseSequenceGenerator,
    bins: int = 500,
    seq_len: int = 1000,
    batch_size: int = 10000,
    num_batches: int = 1,
    rng: int | Generator = 13683624337160779813,
    figsize: tuple[int, int] = (16, 5),
    scale: str = "identity",
    **kwargs,
) -> plt.Figure:
    r"""Plots the distribution from a sequence generator.

    Args:
    ----
        sequence (``BaseSequenceGenerator``): Specifies the sequence
            generator.
        bins (int, optional): Specifies the number of histogram bins.
            Default: ``500``
        seq_len (int, optional): Specifies the sequence length.
            Default: ``128``
        batch_size (int, optional): Specifies the batch size.
            Default: ``1``
        rng (``torch.Generator`` or int): Specifies a random number
            generator or a random seed.
            Default: ``13683624337160779813``
        figsize (tuple, optional): Specifies the figure size.
            Default: ``(16, 5)``
        **kwargs: Additional keyword arguments for ``plt.hist``.

    Returns:
    -------
        ``matplotlib.pyplot.Figure``: The generated figure.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.utils.plot import hist_sequence
        >>> from startorch.sequence import RandUniform
        >>> generator = RandUniform(low=-5, high=5)
        >>> fig = hist_sequence(generator)
    """
    check_matplotlib()
    if not isinstance(rng, Generator):
        rng = get_torch_generator(random_seed=rng)

    batch = merge_batches(
        [
            sequence.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
            for _ in range(num_batches)
        ]
    )
    batch = scale_batch(batch, scale=scale)
    fig, ax = plt.subplots(figsize=figsize)
    ax.hist(batch.data.flatten().numpy(), bins=bins, **kwargs)
    return fig


def plot_sequence(
    sequence: BaseSequenceGenerator,
    seq_len: int = 128,
    batch_size: int = 1,
    num_batches: int = 1,
    rng: int | Generator = 13683624337160779813,
    figsize: tuple[int, int] = (16, 5),
    xscale: str = "linear",
    yscale: str = "linear",
    **kwargs,
) -> plt.Figure:
    r"""Plots some sequences generated from a sequence generator.

    Args:
    ----
        sequence (``BaseSequenceGenerator``): Specifies the sequence
            generator.
        seq_len (int, optional): Specifies the sequence length.
            Default: ``128``
        batch_size (int, optional): Specifies the batch size.
            Default: ``1``
        num_batches (int, optional): Specifies the number of batches.
            Default: ``1``
        rng (``torch.Generator`` or int): Specifies a random number
            generator or a random seed.
            Default: ``13683624337160779813``
        figsize (tuple, optional): Specifies the figure size.
            Default: ``(16, 5)``
        xscale (str, optional): Specifies the x-axis scale.
            Default: ``linear``
        yscale (str, optional): Specifies the y-axis scale.
            Default: ``linear``
        **kwargs: Additional keyword arguments for ``plt.plot``.

    Returns:
    -------
        ``matplotlib.pyplot.Figure``: The generated figure.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.utils.plot import plot_sequence
        >>> from startorch.sequence import RandUniform
        >>> generator = RandUniform(low=-5, high=5)
        >>> fig = plot_sequence(generator, batch_size=4)
    """
    check_matplotlib()
    if not isinstance(rng, Generator):
        rng = get_torch_generator(random_seed=rng)

    fig, ax = plt.subplots(figsize=figsize)
    for _ in range(num_batches):
        batch = sequence.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
        for i in range(batch.batch_size):
            ax.plot(batch.select_along_batch(i).data.flatten().numpy(), marker="o", **kwargs)
    ax.set_xscale(xscale)
    ax.set_yscale(yscale)
    return fig



from __future__ import annotations

__all__ = [
    "BaseTimeSeriesGenerator",
    "MultinomialChoice",
    "Periodic",
    "TimeSeries",
    "is_timeseries_generator_config",
    "merge_multiple_timeseries_by_time",
    "setup_timeseries_generator",
    "truncate_timeseries_by_length",
]

from startorch.timeseries.base import (
    BaseTimeSeriesGenerator,
    is_timeseries_generator_config,
    setup_timeseries_generator,
)
from startorch.timeseries.choice import (
    MultinomialChoiceTimeSeriesGenerator as MultinomialChoice,
)
from startorch.timeseries.generic import TimeSeriesGenerator as TimeSeries
from startorch.timeseries.merge import (
    merge_multiple_timeseries_by_time,
    truncate_timeseries_by_length,
)
from startorch.timeseries.periodic import PeriodicTimeSeriesGenerator as Periodic



from __future__ import annotations

__all__ = [
    "ExponentialSequenceGenerator",
    "RandExponentialSequenceGenerator",
    "RandTruncExponentialSequenceGenerator",
    "TruncExponentialSequenceGenerator",
]


from coola.utils.format import str_indent, str_mapping
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.random import (
    exponential,
    rand_exponential,
    rand_trunc_exponential,
    trunc_exponential,
)
from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator
from startorch.sequence.constant import ConstantSequenceGenerator, FullSequenceGenerator
from startorch.utils.conversion import to_tuple


class ExponentialSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values from
    an Exponential distribution.

    The rates of the Exponential distribution are generated by the
    rate generator. The rate generator should return the rate for each
    value in the sequence.

    Args:
    ----
        rate (``BaseSequenceGenerator`` or dict):
            Specifies the rate generator or its configuration.
            The rate generator should return valid rate values.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import Exponential, RandUniform
        >>> generator = Exponential(rate=RandUniform(low=1.0, high=10.0))
        >>> generator
        ExponentialSequenceGenerator(
          (rate): RandUniformSequenceGenerator(low=1.0, high=10.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(self, rate: BaseSequenceGenerator | dict) -> None:
        super().__init__()
        self._rate = setup_sequence_generator(rate)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"rate": self._rate}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            exponential(
                self._rate.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                generator=rng,
            )
        )

    @classmethod
    def create_fixed_rate(
        cls, rate: float = 1.0, feature_size: tuple[int, ...] | list[int] | int = 1
    ) -> ExponentialSequenceGenerator:
        r"""Implements a sequence generator where the values are sampled
        from an Exponential distribution with a fixed rate.

        Args:
        ----
            rate (float, optional): Specifies the rate of the
                Exponential distribution. Default: ``1.0``
            feature_size (tuple or list or int, optional): Specifies the
                feature size. Default: ``1``

        Returns:
        -------
            ``ExponentialSequenceGenerator``: A sequence generator where
                the rates of the Exponential distribution are a fixed
                given value.

        Example usage:

        .. code-block:: pycon

            >>> from startorch.sequence import Exponential, RandUniform
            >>> generator = Exponential.create_fixed_rate(rate=1.0)
            >>> generator
            ExponentialSequenceGenerator(
              (rate): FullSequenceGenerator(value=1.0, feature_size=(1,))
            )
            >>> generator.generate(seq_len=6, batch_size=2)
            tensor([[...]], batch_dim=0, seq_dim=1)
        """
        return cls(FullSequenceGenerator(value=rate, feature_size=feature_size))

    @classmethod
    def create_uniform_rate(
        cls,
        min_rate: float = 0.01,
        max_rate: float = 1.0,
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> ExponentialSequenceGenerator:
        r"""Implements a sequence generator where the rates of the
        Exponential distribution are sampled from a uniform
        distribution.

        One rate is sampled per sequence.

        Args:
        ----
            min_rate (float, optional): Specifies the minimum rate
                value. Default: ``0.01``
            max_rate (float, optional): Specifies the maximum rate
                value. Default: ``1.0``
            feature_size (tuple or list or int, optional): Specifies the
                feature size. Default: ``1``

        Returns:
        -------
            ``ExponentialSequenceGenerator``: A sequence generator where
                the rates for each sequence are sampled from a
                uniform distribution.

        Example usage:

        .. code-block:: pycon

            >>> from startorch.sequence import Exponential, RandUniform
            >>> generator = Exponential.create_uniform_rate(min_rate=0.1, max_rate=1.0)
            >>> generator
            ExponentialSequenceGenerator(
              (rate): ConstantSequenceGenerator(
                  (sequence): RandUniformSequenceGenerator(low=0.1, high=1.0, feature_size=(1,))
                )
            )
            >>> generator.generate(seq_len=6, batch_size=2)
            tensor([[...]], batch_dim=0, seq_dim=1)
        """
        # The import is here to do not generate circular dependencies
        from startorch.sequence.uniform import RandUniformSequenceGenerator

        return cls(
            ConstantSequenceGenerator(
                RandUniformSequenceGenerator(
                    low=min_rate,
                    high=max_rate,
                    feature_size=feature_size,
                )
            ),
        )


class RandExponentialSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequences by sampling values from
    an Exponential distribution.

    Args:
    ----
        rate (float, optional): Specifies the rate of the Exponential
            distribution. Default: ``1.0``
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Raises:
    ------
        ValueError if ``rate`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import RandExponential
        >>> generator = RandExponential(rate=1.0)
        >>> generator
        RandExponentialSequenceGenerator(rate=1.0, feature_size=(1,))
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        rate: float = 1.0,
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> None:
        super().__init__()
        if rate <= 0:
            raise ValueError(f"rate has to be greater than 0 (received: {rate})")
        self._rate = float(rate)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(rate={self._rate}, "
            f"feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            rand_exponential(
                size=(batch_size, seq_len) + self._feature_size,
                rate=self._rate,
                generator=rng,
            )
        )


class RandTruncExponentialSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequences by sampling values from
    a truncated Exponential distribution.

    Args:
    ----
        rate (float, optional): Specifies the rate of the Exponential
            distribution. Default: ``1.0``
        max_value (float, optional): Specifies the maximum value.
            Default: ``5.0``
        feature_size (tuple or list or int, optional): Specifies the
            feature size. Default: ``1``

    Raises:
    ------
        ValueError if ``rate`` is not a positive number.
        ValueError if ``max_value`` is not a positive number.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import RandTruncExponential
        >>> generator = RandTruncExponential(rate=1.0, max_value=3.0)
        >>> generator
        RandTruncExponentialSequenceGenerator(rate=1.0, max_value=3.0, feature_size=(1,))
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        rate: float = 1.0,
        max_value: float = 5.0,
        feature_size: tuple[int, ...] | list[int] | int = 1,
    ) -> None:
        super().__init__()
        if rate <= 0:
            raise ValueError(f"rate has to be greater than 0 (received: {rate})")
        self._rate = float(rate)
        if max_value <= 0:
            raise ValueError(f"max_value has to be greater than 0 (received: {max_value})")
        self._max_value = float(max_value)
        self._feature_size = to_tuple(feature_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(rate={self._rate}, max_value={self._max_value}, "
            f"feature_size={self._feature_size})"
        )

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            rand_trunc_exponential(
                size=(batch_size, seq_len) + self._feature_size,
                rate=self._rate,
                max_value=self._max_value,
                generator=rng,
            )
        )


class TruncExponentialSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values from
    an Exponential distribution.

    Args:
    ----
        rate (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the rate.
        max_value (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            maximum value (excluded).

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import RandUniform, TruncExponential
        >>> generator = TruncExponential(
        ...     rate=RandUniform(low=1.0, high=10.0),
        ...     max_value=RandUniform(low=1.0, high=100.0),
        ... )
        >>> generator
        TruncExponentialSequenceGenerator(
          (rate): RandUniformSequenceGenerator(low=1.0, high=10.0, feature_size=(1,))
          (max_value): RandUniformSequenceGenerator(low=1.0, high=100.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        rate: BaseSequenceGenerator | dict,
        max_value: BaseSequenceGenerator | dict,
    ) -> None:
        super().__init__()
        self._rate = setup_sequence_generator(rate)
        self._max_value = setup_sequence_generator(max_value)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"rate": self._rate, "max_value": self._max_value}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return BatchedTensorSeq(
            trunc_exponential(
                rate=self._rate.generate(seq_len=seq_len, batch_size=batch_size, rng=rng).data,
                max_value=self._max_value.generate(
                    seq_len=seq_len, batch_size=batch_size, rng=rng
                ).data,
                generator=rng,
            )
        )



from __future__ import annotations

__all__ = ["shapes_are_equal"]

from collections.abc import Sequence

from torch import Tensor


def shapes_are_equal(tensors: Sequence[Tensor]) -> bool:
    r"""Indicates if the shapes of several tensors are equal or not.

    This method does not check the values or the data type of the
    tensors.

    Args:
    ----
        tensors (sequence): Specifies the tensors to check.

    Returns:
    -------
        bool: ``True`` if all the tensors have the same shape,
            otherwise ``False``. By design, this function returns
            ``False`` if no tensor is provided.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.utils.tensor import shapes_are_equal
        >>> shapes_are_equal([torch.rand(2, 3), torch.rand(2, 3)])
        True
        >>> shapes_are_equal([torch.rand(2, 3), torch.rand(2, 3, 1)])
        False
    """
    if not tensors:
        return False
    shape = tensors[0].shape
    return all(shape == tensor.shape for tensor in tensors[1:])



from __future__ import annotations

__all__ = ["str_target_object", "str_weighted_modules"]

from collections.abc import Sequence

from coola.utils import str_indent
from objectory import OBJECT_TARGET


def str_target_object(config: dict) -> str:
    r"""Gets a string that indicates the target object in the config.

    Args:
    ----
        config (dict): Specifies a config using the ``object_factory``
            library. This dict is expected to have a key
            ``'_target_'`` to indicate the target object.

    Returns:
    -------
        str: A string with the target object.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.utils.format import str_target_object
        >>> str_target_object({OBJECT_TARGET: "something.MyClass"})
        [_target_: something.MyClass]
        >>> str_target_object({})
        [_target_: N/A]
    """
    return f"[{OBJECT_TARGET}: {config.get(OBJECT_TARGET, 'N/A')}]"


def str_weighted_modules(modules: Sequence, weights: Sequence, num_spaces: int = 2) -> str:
    r"""Computes a pretty representation of a sequence of modules where
    each module is associated to a weight.

    Args:
        modules (``Sequence``): Specifies the modules.
        weights (``Sequence``): Specifies the weights. The ``weights``
            should have the same length that ``modules``.
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.

    Returns:
        str: The string representation of the modules.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.utils.format import str_weighted_modules
        >>> print(str_weighted_modules(modules=["abc", "something\nelse"], weights=[1, 2]))
        (0) [weight=1] abc
        (1) [weight=2] something
          else
    """
    if len(modules) != len(weights):
        raise RuntimeError(
            f"`modules` and `weights` must have the same length but received {len(modules):,} "
            f"and {len(weights):,}"
        )

    lines = []
    for i, (item, weight) in enumerate(zip(modules, weights)):
        lines.append(f"({i}) [weight={weight}] {str_indent(str(item), num_spaces=num_spaces)}")
    return "\n".join(lines)



from __future__ import annotations

__all__ = ["LinearSequenceGenerator"]


from coola.utils import str_indent, str_mapping
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.sequence.base import BaseSequenceGenerator, setup_sequence_generator


class LinearSequenceGenerator(BaseSequenceGenerator):
    r"""Implements a class to generate sequence by sampling values with a
    linear pattern.

    The sequences are generated by using the following formula:

    ``output = slope * value + intercept``

    Args:
    ----
        value (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            sequence values. The linear transformation is applied
            on these values.
        slope (``BaseSequenceGenerator`` or dict): Specifies a sequence
            generator (or its configuration) to generate the slope
            values.
        intercept (``BaseSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            intercept values.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.sequence import Linear, RandUniform
        >>> generator = Linear(
        ...     value=RandUniform(low=-1.0, high=1.0),
        ...     slope=RandUniform(low=1.0, high=2.0),
        ...     intercept=RandUniform(low=-10.0, high=-5.0),
        ... )
        >>> generator
        LinearSequenceGenerator(
          (value): RandUniformSequenceGenerator(low=-1.0, high=1.0, feature_size=(1,))
          (slope): RandUniformSequenceGenerator(low=1.0, high=2.0, feature_size=(1,))
          (intercept): RandUniformSequenceGenerator(low=-10.0, high=-5.0, feature_size=(1,))
        )
        >>> generator.generate(seq_len=6, batch_size=2)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        value: BaseSequenceGenerator | dict,
        slope: BaseSequenceGenerator | dict,
        intercept: BaseSequenceGenerator | dict,
    ) -> None:
        super().__init__()
        self._value = setup_sequence_generator(value)
        self._slope = setup_sequence_generator(slope)
        self._intercept = setup_sequence_generator(intercept)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping({"value": self._value, "slope": self._slope, "intercept": self._intercept})
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return (
            self._value.generate(seq_len=seq_len, batch_size=batch_size, rng=rng)
            .mul(self._slope.generate(seq_len=seq_len, batch_size=batch_size, rng=rng))
            .add(self._intercept.generate(seq_len=seq_len, batch_size=batch_size, rng=rng))
        )



__all__ = [
    "BATCH_SIZE",
    "FEATURE",
    "SEQUENCE",
    "SEQ_LEN",
    "TARGET",
    "TIME",
    "VALUE",
]

BATCH_SIZE = "batch_size"
FEATURE = "feature"
SEQUENCE = "sequence"
SEQ_LEN = "seq_len"
TARGET = "target"
TIME = "time"
VALUE = "value"



from __future__ import annotations

__all__ = [
    "AcoshTensorGenerator",
    "AsinhTensorGenerator",
    "AtanhTensorGenerator",
    "CoshTensorGenerator",
    "SinhTensorGenerator",
    "TanhTensorGenerator",
]

from torch import Generator, Tensor

from startorch.tensor.wrapper import BaseWrapperTensorGenerator


class AcoshTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a sequence generator that computes the inverse
    hyperbolic cosine (arccosh) of each value.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Acosh, RandUniform
        >>> generator = Acosh(RandUniform(low=1.0, high=5.0))
        >>> generator
        AcoshTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=1.0, high=5.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).acosh()


class AsinhTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a sequence generator that computes the inverse
    hyperbolic sine (arcsinh) of each value.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Asinh, RandUniform
        >>> generator = Asinh(RandUniform(low=0.0, high=1000.0))
        >>> generator
        AsinhTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=0.0, high=1000.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).asinh()


class AtanhTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a sequence generator that computes the inverse
    hyperbolic tangent (arctanh) of each value.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Atanh, RandUniform
        >>> generator = Atanh(RandUniform(low=-0.5, high=0.5))
        >>> generator
        AtanhTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=-0.5, high=0.5)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).atanh()


class CoshTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a sequence generator that computes the hyperbolic
    cosine (cosh) of each value.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import Cosh, RandUniform
        >>> generator = Cosh(RandUniform())
        >>> generator
        CoshTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=0.0, high=1.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).cosh()


class SinhTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a sequence generator that computes the hyperbolic sine
    (sinh) of each value.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform, Sinh
        >>> generator = Sinh(RandUniform(low=0.0, high=1.0))
        >>> generator
        SinhTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=0.0, high=1.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).sinh()


class TanhTensorGenerator(BaseWrapperTensorGenerator):
    r"""Implements a sequence generator that computes the hyperbolic
    tangent (tanh) of each value.

    Example usage:

    .. code-block:: pycon

        >>> from startorch.tensor import RandUniform, Tanh
        >>> generator = Tanh(RandUniform(low=0.0, high=1.0))
        >>> generator
        TanhTensorGenerator(
          (tensor): RandUniformTensorGenerator(low=0.0, high=1.0)
        )
        >>> generator.generate((2, 6))
        tensor([[...]])
    """

    def generate(self, size: tuple[int, ...], rng: Generator | None = None) -> Tensor:
        return self._generator.generate(size=size, rng=rng).tanh()



from __future__ import annotations

__all__ = ["SineWavePeriodicSequenceGenerator"]

import math

from coola.utils.format import str_indent, str_mapping
from redcat import BatchedTensorSeq
from torch import Generator

from startorch.periodic.sequence import (
    BasePeriodicSequenceGenerator,
    setup_periodic_sequence_generator,
)


class SineWavePeriodicSequenceGenerator(BasePeriodicSequenceGenerator):
    r"""Implements a class to generate periodic sequence by sampling
    values with a sine wave pattern.

    The sequences are generated by using the following formula:

    ``output = amplitude * sin(2 * pi * frequency * value + phase)``

    Args:
    ----
        value (``BasePeriodicSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            sequence values.
        phase (``BasePeriodicSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            phase values.
        amplitude (``BasePeriodicSequenceGenerator`` or dict): Specifies a
            sequence generator (or its configuration) to generate the
            amplitude values.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from startorch.periodic.sequence import Repeat, SineWave
        >>> from startorch.sequence import RandUniform
        >>> generator = SineWave(
        ...     value=Repeat(RandUniform(low=-1.0, high=1.0)),
        ...     phase=Repeat(RandUniform(low=-1.0, high=1.0)),
        ...     amplitude=Repeat(RandUniform(low=-1.0, high=1.0)),
        ... )
        >>> generator
        SineWavePeriodicSequenceGenerator(
          (value): RepeatPeriodicSequenceGenerator(
              (generator): RandUniformSequenceGenerator(low=-1.0, high=1.0, feature_size=(1,))
            )
          (phase): RepeatPeriodicSequenceGenerator(
              (generator): RandUniformSequenceGenerator(low=-1.0, high=1.0, feature_size=(1,))
            )
          (amplitude): RepeatPeriodicSequenceGenerator(
              (generator): RandUniformSequenceGenerator(low=-1.0, high=1.0, feature_size=(1,))
            )
        )
        >>> generator.generate(seq_len=12, period=4, batch_size=4)
        tensor([[...]], batch_dim=0, seq_dim=1)
    """

    def __init__(
        self,
        value: BasePeriodicSequenceGenerator | dict,
        phase: BasePeriodicSequenceGenerator | dict,
        amplitude: BasePeriodicSequenceGenerator | dict,
    ) -> None:
        super().__init__()
        self._value = setup_periodic_sequence_generator(value)
        self._phase = setup_periodic_sequence_generator(phase)
        self._amplitude = setup_periodic_sequence_generator(amplitude)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping({"value": self._value, "phase": self._phase, "amplitude": self._amplitude})
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(
        self, seq_len: int, period: int, batch_size: int = 1, rng: Generator | None = None
    ) -> BatchedTensorSeq:
        return (
            self._value.generate(seq_len=seq_len, period=period, batch_size=batch_size, rng=rng)
            .mul(2.0 * math.pi / period)
            .add(
                self._phase.generate(seq_len=seq_len, period=period, batch_size=batch_size, rng=rng)
            )
            .sin()
            .mul(
                self._amplitude.generate(
                    seq_len=seq_len, period=period, batch_size=batch_size, rng=rng
                )
            )
        )
